import itertools

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import scipy.spatial.distance as distance
import seaborn as sns

from immunedb.common.models import CloneStats, Sample
from immunedb.exporting.writer import ExportWriter
from immunedb.util.log import logger


def get_feature_str(sample, features):
    if 'sample' in features:
        return sample.name
    if 'subject' in features:
        return sample.subject.identifier
    s = ' '.join((sample.metadata_dict.get(f, 'NA') for f in features))
    return s


def get_sample_df(session, sample_ids, features, size_metric, dist_func):
    size_metric = {
        'copies': 'total_cnt',
        'instances': 'unique_cnt'
    }[size_metric]
    stats = session.query(
        CloneStats.sample_id,
        CloneStats.clone_id,
        getattr(CloneStats, size_metric).label('size')
    ).filter(
        CloneStats.sample_id.in_(sample_ids),
    )

    clone_sizes = {}
    for stat in stats:
        clone_sizes.setdefault(stat.clone_id, {})[stat.sample_id] = stat.size
    clone_sizes = pd.DataFrame(clone_sizes).fillna(0)

    sim = {}
    for s1, s2 in list(itertools.combinations(clone_sizes.index, 2)):
        fsim = 1 - dist_func(clone_sizes.loc[s1], clone_sizes.loc[s2])
        sim.setdefault(s1, {})[s2] = sim.setdefault(s2, {})[s1] = fsim

    return pd.DataFrame(sim)


def collapse_df_features(df, features, sample_instances, agg_func):
    def _apply_feature(arr):
        return [
            get_feature_str(sample_instances[sample_id], features)
            for sample_id in arr
        ]
    df.columns = _apply_feature(df.columns)
    df.index = _apply_feature(df.index)
    df = df.groupby(
        df.columns, axis=1
    ).agg(agg_func).groupby(
        df.index, axis=0
    ).agg(agg_func)
    return df.reindex(sorted(df.columns), axis=1).sort_index()


def write_clone_overlap(session, sample_ids=None, pool_on=('sample',),
                        size_metric='copies', sim_func='cosine',
                        agg_func='median', zipped=False, **kwargs):
    samples = session.query(Sample)
    if sample_ids:
        samples = samples.filter(Sample.id.in_(sample_ids))
    sample_instances = {s.id: s for s in samples}

    with ExportWriter(zipped=zipped) as writer:
        for subject in set([s.subject for s in sample_instances.values()]):
            logger.info('Calculating overlap for {}'.format(
                subject.identifier))
            sub_samples = [
                s.id for s in sample_instances.values() if s.subject == subject
            ]
            sdf = get_sample_df(session, sub_samples, pool_on, size_metric,
                                getattr(distance, sim_func))
            if sdf.empty:
                logger.warning(
                    'Subject {} had no clones for calculation'.format(
                        subject.identifier))
                continue

            sdf = collapse_df_features(sdf, pool_on, sample_instances,
                                       getattr(np, agg_func))
            name = '{}.overlap'.format(subject.identifier)

            with writer.get_handle(name + '.tsv') as fh:
                sdf.to_csv(fh, sep='\t')

            title_fmt = 'Subject {}\npooled by={}, similarity metric={}'
            if 'sample' not in pool_on:
                title_fmt += ', aggregation function={}'
            fig, ax = plt.subplots(figsize=(20, 20))
            ax = sns.heatmap(sdf, annot=True, linewidths=.25, vmin=0, vmax=1)
            ax.set_title(title_fmt.format(
                subject.identifier, ' & '.join(pool_on), sim_func, agg_func
            ))

            with writer.get_handle(name + '.pdf', 'wb+') as fh:
                plt.savefig(fh, bbox_inches='tight', format='pdf')

        return writer.get_zip_value()



from collections import OrderedDict
from sqlalchemy.orm import joinedload

from immunedb.identification.genes import GeneName
from immunedb.exporting.writer import ExportWriter
from immunedb.common.models import (Sample, SampleMetadata, Sequence,
                                    SequenceCollapse)
from immunedb.util.log import logger
from immunedb.util.funcs import yield_limit
from immunedb.exporting.tsv_writer import StreamingTSV


mappings = {
    'airr': OrderedDict((
        ('sequence_id', 'seq_id'),
        ('sequence', 'original_sequence'),
        ('rev_comp', 'rev_comp'),
        ('productive', 'functional'),
        ('vj_in_frame', 'in_frame'),
        ('stop_codon', 'stop'),
        ('locus', lambda s: GeneName(s.v_gene).base),
        ('v_call', 'v_gene'),
        ('d_call', None),
        ('j_call', 'j_gene'),
        ('sequence_alignment', 'sequence'),
        ('germline_alignment', 'germline'),
        ('junction', 'cdr3_nt'),
        ('junction_aa', 'cdr3_aa'),
        ('v_score', 'v_match'),
        ('v_identity', lambda s: s.v_match / s.v_length),
        ('v_cigar', 'v_cigar'),
        ('d_cigar', None),
        ('j_identity', lambda s: s.j_match / s.j_length),
        ('j_score', 'j_match'),
        ('j_cigar', 'j_cigar'),
        ('v_sequence_start', 'seq_start'),
        ('duplicate_count', 'copy_number'),
        ('clone_id', 'clone_id'),

        # Extras
        ('clone_junction_nt', lambda s: s.clone.cdr3_nt if s.clone else ''),
        ('clone_junction_aa', lambda s: s.clone.cdr3_aa if s.clone else ''),
        ('sample', lambda s: s.sample.name),
        ('subject', lambda s: s.subject.identifier),
        ('duplicate_count_in_subject', lambda s:
            s.collapse.copy_number_in_subject))),


    'changeo': OrderedDict((
        ('SEQUENCE_ID', 'seq_id'),
        ('SEQUENCE_IMGT', 'sequence'),
        ('GERMLINE_IMGT_D_MASK', 'germline_d_masked'),
        ('FUNCTIONAL', 'functional'),
        ('IN_FRAME', 'in_frame'),
        ('STOP', 'stop'),
        ('V_CALL', lambda s: s.v_gene.replace(
           '|', ',' + GeneName(s.v_gene).prefix)),
        ('J_CALL', lambda s: s.j_gene.replace(
           '|', ',' + GeneName(s.j_gene).prefix)),
        ('JUNCTION_LENGTH', 'cdr3_num_nts'),
        ('JUNCTION', 'cdr3_nt'),
        ('V_IDENTITY', lambda s: round(100 * s.v_match / s.v_length, 2)),
        ('V_SCORE', 'v_match'),
        ('J_IDENTITY', lambda s: round(100 * s.j_match / s.j_length, 2)),
        ('J_SCORE', 'j_match'),
        ('DUPCOUNT', lambda s: s.copy_number),
        ('DUPCOUNT_IN_SUBJECT', lambda s: s.collapse.copy_number_in_subject),
        ('CLONE', 'clone_id'),

        # Extras
        ('CLONE_CDR3_NT', lambda s: s.clone.cdr3_nt if s.clone else ''),
        ('CLONE_CDR3_AA', lambda s: s.clone.cdr3_aa if s.clone else ''),
        ('SAMPLE', lambda s: s.sample.name),
        ('SUBJECT', lambda s: s.subject.identifier)))
}


class SequenceWriter(StreamingTSV):
    META_PREFIX = 'METADATA_'

    def __init__(self, format_name, metadata_fields):
        self.mapping = mappings[format_name]
        fields = list(self.mapping.keys()) + list(sorted([
            self.META_PREFIX + m for m in metadata_fields]))
        super(SequenceWriter, self).__init__(fields)

    def writeseq(self, seq):
        return super(SequenceWriter, self).writerow(self.format_seq(seq))

    def format_seq(self, seq):
        fields = {
            out_field: self._get_val(seq, model_field)
            for out_field, model_field in self.mapping.items()
        }

        fields.update({
            self.META_PREFIX + k: v
            for k, v in seq.sample.metadata_dict.items()
        })
        return fields

    def _get_val(self, model, field):
        if type(field) == str:
            val = getattr(model, field)
            if type(val) == bool:
                return 'T' if val else 'F'
            return val
        if callable(field):
            return field(model)
        return ''


def get_sequences(session, sample, fmt, clones_only, min_subject_copies):
    meta_keys = set([m.key for m in session.query(SampleMetadata.key)])

    seqs = session.query(Sequence).filter(
        Sequence.sample_id == sample.id
    ).join(
        SequenceCollapse
    ).options(
        joinedload(Sequence.clone),
        joinedload(Sequence.collapse),
        joinedload(Sequence.sample),
        joinedload(Sequence.subject),
    )

    if clones_only:
        seqs = seqs.filter(~Sequence.clone_id.is_(None))
    if min_subject_copies:
        seqs = seqs.filter(
            SequenceCollapse.copy_number_in_subject >=
            min_subject_copies
        )

    writer = SequenceWriter(fmt, meta_keys)

    yield writer.writeheader()
    for seq in yield_limit(seqs, Sequence.ai):
        yield writer.writeseq(seq)


def write_sequences(session, sample_ids=None, out_format='changeo',
                    clones_only=False, min_subject_copies=None, zipped=False,
                    **kwargs):
    samples = session.query(Sample)
    if sample_ids:
        samples = samples.filter(Sample.id.in_(sample_ids))
    with ExportWriter(zipped=zipped) as fh:
        for sample in samples:
            logger.info('Exporting sample {}'.format(sample.name))
            fh.set_filename('{}.{}.tsv'.format(sample.name, out_format))
            fh.write(
                get_sequences(session, sample, out_format, clones_only,
                              min_subject_copies)
            )
        return fh.get_zip_value()



import re

import dnautils

from immunedb.common.models import CDR3_OFFSET
from immunedb.identification import AlignmentException, get_common_seq
from immunedb.identification.genes import VGene, find_v_position
from immunedb.identification.vdj_sequence import VDJAlignment


def sliding_window_match(sequence, match):
    r = re.search(match.replace('N', '.'), sequence)
    return r.start() if r else -1


class AnchorAligner(object):
    MISMATCH_THRESHOLD = 3

    def __init__(self, v_germlines, j_germlines):
        self.v_germlines = v_germlines
        self.j_germlines = j_germlines

    def get_alignment(self, vdj_sequence, limit_vs=None, limit_js=None):
        alignment = VDJAlignment(vdj_sequence)
        self.find_j(alignment, limit_js)
        self.find_v(alignment, limit_vs)
        return alignment

    def _find_index(self, sequence, germline):
        best_pos, best_hamming = None, None
        for pos in range(len(sequence) - len(germline)):
            hamming = dnautils.hamming(sequence[pos:pos + len(germline)],
                                       germline) / len(germline)
            if best_hamming is None or hamming < best_hamming:
                best_pos = pos
                best_hamming = hamming
                is_rc = False

        rc = sequence.reverse_complement()
        for pos in range(len(rc) - len(germline) + 1):
            hamming = dnautils.hamming(rc[pos:pos + len(germline)],
                                       germline) / len(germline)
            if best_hamming is None or hamming < best_hamming:
                best_pos = pos
                best_hamming = hamming
                is_rc = True

        best_pos += len(germline) - self.j_germlines.anchor_len
        return best_pos, best_hamming, is_rc

    def find_j(self, alignment, limit_js):
        # Iterate over every possible J anchor.  For each germline, try its
        # full sequence, then exclude the final 3 characters at a time until
        # there are only MIN_J_ANCHOR_LEN nucleotides remaining.
        #
        # For example, the order for one germline:
        # TGGTCACCGTCTCCTCAG
        # TGGTCACCGTCTCCT
        # TGGTCACCGTCT

        rc = alignment.sequence.reverse_complement()
        for match, j_gene in self.j_germlines.get_all_anchors(limit_js):
            i = alignment.sequence.rfind(match)
            if i >= 0:
                return self.process_j(alignment, i, len(match), limit_js)

            i = rc.rfind(match)
            if i >= 0:
                alignment.sequence = rc
                return self.process_j(alignment, i, len(match), limit_js)

            i = sliding_window_match(alignment.sequence.sequence, match)
            if i >= 0:
                return self.process_j(alignment, i, len(match), limit_js)

            i = sliding_window_match(rc.sequence, match)
            if i >= 0:
                alignment.sequence = rc
                return self.process_j(alignment, i, len(match), limit_js)

        # Last chance, find any matching position
        total_best_hamming = None
        total_best_rc, total_best_pos = None, None
        for j_gene, germ_seq in self.j_germlines.items():
            best_pos, best_hamming, is_rc = self._find_index(
                alignment.sequence, germ_seq)
            if (total_best_hamming is None or
                    best_hamming < total_best_hamming):
                total_best_hamming = best_hamming
                total_best_pos = best_pos
                total_best_rc = is_rc

        if total_best_rc:
            alignment.sequence = rc

        return self.process_j(alignment, total_best_pos,
                              self.j_germlines.anchor_len, limit_js)

    def process_j(self, alignment, i, match_len, limit_js):
        # If a match is found, record its location and gene
        alignment.j_anchor_pos = i
        end_of_j = alignment.j_anchor_pos + self.j_germlines.anchor_len
        if end_of_j > len(alignment.sequence):
            alignment.sequence.pad_right(end_of_j - len(alignment.sequence))
        if limit_js:
            j_germs = {
                k: v for k, v in self.j_germlines.items()
                if k.name in limit_js
            }
        else:
            j_germs = self.j_germlines

        best_dist = None
        for j_gene, j_seq in j_germs.items():
            seq_j = alignment.sequence[end_of_j - len(j_seq):end_of_j]
            if seq_j:
                dist = dnautils.hamming(seq_j, j_seq[:len(seq_j)]) / len(seq_j)
                if best_dist is None or dist < best_dist:
                    best_dist = dist
                    alignment.j_gene = set([j_gene])
                elif dist == best_dist:
                    alignment.j_gene.add(j_gene)

        if len(alignment.j_gene) == 0:
            raise AlignmentException('Could not find suitable J anchor')

        # Get the full germline J gene
        ex_j = sorted(alignment.j_gene)[0]
        j_full = self.j_germlines[ex_j]

        # Get the portion of the germline J in the CDR3
        germline_in_cdr3 = self.j_germlines.get_j_in_cdr3(ex_j)
        cdr3_end_pos = (
            alignment.j_anchor_pos + self.j_germlines.anchor_len -
            self.j_germlines.upstream_of_cdr3
        )
        sequence_in_cdr3 = alignment.sequence[
            cdr3_end_pos - len(germline_in_cdr3):
            cdr3_end_pos
        ]
        if len(germline_in_cdr3) == 0 or len(sequence_in_cdr3) == 0:
            alignment.j_gene = set()
            raise AlignmentException('Could not find sequence or germline in '
                                     'CDR3')

        # Get the extent of the J in the CDR3
        streak = dnautils.find_streak_position(
            germline_in_cdr3[::-1],
            sequence_in_cdr3[::-1],
            self.MISMATCH_THRESHOLD)

        # Trim the J gene based on the extent in the CDR3
        if streak is not None:
            j_full = j_full[len(germline_in_cdr3) - streak:]
            alignment.germline_cdr3 = germline_in_cdr3[-streak:]
        else:
            alignment.germline_cdr3 = germline_in_cdr3

        # Find where the full J starts
        j_start = alignment.j_anchor_pos + match_len - len(j_full)

        # If the trimmed germline J extends past the end of the
        # sequence, there is a misalignment
        if len(j_full) != len(alignment.sequence[j_start:j_start+len(j_full)]):
            alignment.j_gene = set()
            raise AlignmentException('Germline extended past end of J')

        alignment.j_length = len(j_full)
        alignment.post_cdr3_length = self.j_germlines.upstream_of_cdr3

    def find_v(self, alignment, limit_vs):
        for anchor_pos in find_v_position(alignment.sequence.sequence):
            self.process_v(alignment, anchor_pos, limit_vs)
            if len(alignment.v_gene) > 0:
                break

        if len(alignment.v_gene) == 0:
            raise AlignmentException('Could not find suitable V anchor')

    def process_v(self, alignment, anchor_pos, limit_vs):
        aligned_v = VGene(alignment.sequence.sequence)
        v_score = None
        for v, germ in self.v_germlines.alignments.items():
            if limit_vs is not None and v.name not in limit_vs:
                continue
            try:
                dist, total_length = germ.compare(aligned_v,
                                                  alignment.j_anchor_pos,
                                                  self.MISMATCH_THRESHOLD)
            except Exception:
                continue
            # Record this germline if it is has the lowest distance
            if dist is not None:
                if v_score is None or dist < v_score:
                    alignment.v_gene = set([v])
                    alignment.v_length = total_length
                    germ_pos = germ.ungapped_anchor_pos
                    v_score = dist
                elif dist == v_score:
                    # Add the V-tie
                    alignment.v_gene.add(v)

        if len(alignment.v_gene) > 0:
            # Determine the pad length
            alignment.seq_offset = germ_pos - anchor_pos
            # Mutation ratio is the distance divided by the length of overlap
            alignment.v_mutation_fraction = v_score / alignment.v_length

    def align_to_germline(self, alignment, avg_len=None, avg_mut=None):
        if avg_len is not None and avg_mut is not None:
            alignment.v_gene = self.v_germlines.get_ties(
                alignment.v_gene, avg_len, avg_mut)
            alignment.j_gene = self.j_germlines.get_ties(
                alignment.j_gene, avg_len, avg_mut)
        # Set the germline to the V gene up to the CDR3
        germ = get_common_seq(
            [self.v_germlines[v] for v in alignment.v_gene], cutoff=False
        )
        alignment.germline = germ[:CDR3_OFFSET]
        # If we need to pad the sequence, do so, otherwise trim the sequence to
        # the germline length
        if alignment.seq_offset >= 0:
            alignment.sequence.pad(alignment.seq_offset)
        else:
            alignment.sequence.remove_prefix(-alignment.seq_offset)
        alignment.j_anchor_pos += alignment.seq_offset

        # Add germline gaps to sequence before CDR3 and update anchor positions
        for i, c in enumerate(alignment.germline):
            if c == '-':
                alignment.sequence.add_gap(i)
                alignment.j_anchor_pos += 1
                if i < alignment.seq_start:
                    alignment.seq_offset += 1

        j_germ = get_common_seq(
            [self.j_germlines[j] for j in alignment.j_gene], right=True
        )
        # Calculate the length of the CDR3
        alignment.cdr3_num_nts = (
            alignment.j_anchor_pos + self.j_germlines.anchor_len -
            self.j_germlines.upstream_of_cdr3 - alignment.cdr3_start
        )

        v_end = alignment.seq_start + alignment.num_gaps + alignment.v_length
        v_germ = germ[CDR3_OFFSET:v_end]
        alignment.germline_cdr3 = ''.join((
            v_germ,
            '-' * (alignment.cdr3_num_nts -
                   len(v_germ) -
                   len(alignment.germline_cdr3)),
            alignment.germline_cdr3
        ))

        if alignment.cdr3_num_nts < 3:
            raise AlignmentException('CDR3 has no AAs')

        alignment.j_anchor_pos += alignment.cdr3_num_nts
        # Fill germline CDR3 with gaps
        alignment.germline += '-' * alignment.cdr3_num_nts
        alignment.germline += j_germ[-self.j_germlines.upstream_of_cdr3:]
        # If the sequence is longer than the germline, trim it
        if len(alignment.sequence) > len(alignment.germline):
            alignment.sequence.trim_right(len(alignment.germline))
        elif len(alignment.sequence) < len(alignment.germline):
            alignment.sequence.pad_right(len(alignment.germline) -
                                         len(alignment.sequence))
        if len(alignment.cdr3) != alignment.cdr3_num_nts:
            raise AlignmentException('Invalid CDR3 length')



import json

import ete3

from immunedb.common.models import Clone, Sequence, SequenceCollapse
import immunedb.util.concurrent as concurrent
from immunedb.util.log import logger


class LineageWorker(concurrent.Worker):
    def __init__(self, session, newick_generator, min_mut_copies,
                 min_mut_samples, min_seq_copies, min_seq_samples,
                 exclude_stops, full_seq, post_tree_hook=None):
        self.session = session
        self.newick_generator = newick_generator
        self.min_mut_copies = min_mut_copies
        self.min_mut_samples = min_mut_samples
        self.min_seq_copies = min_seq_copies
        self.min_seq_samples = min_seq_samples
        self.exclude_stops = exclude_stops
        self.full_seq = full_seq
        self.post_tree_hook = post_tree_hook

    def get_tree(self, clone, sequences):
        fasta, removed_muts = get_fasta_input(
            clone.consensus_germline, sequences,
            self.min_mut_copies, self.min_mut_samples,
            None if self.full_seq else clone.cdr3_start
        )
        newick = self.newick_generator(fasta)
        if not newick:
            return None
        tree = add_tree_metadata(self.session, newick,
                                 clone.consensus_germline, removed_muts,
                                 None if self.full_seq else clone.cdr3_start)
        if self.post_tree_hook:
            tree = self.post_tree_hook(tree)
        return tree

    def do_task(self, clone_id):
        clone_inst = self.session.query(Clone).filter(
            Clone.id == clone_id).first()
        if not clone_inst:
            return

        self.info('Running clone {}'.format(clone_inst.id))

        sequences = self.session.query(
            Sequence
        ).join(SequenceCollapse).filter(
            Sequence.clone_id == clone_id,
            SequenceCollapse.copy_number_in_subject >= self.min_seq_copies,
            SequenceCollapse.samples_in_subject >= self.min_seq_samples,
        )

        if self.exclude_stops:
            sequences = sequences.filter(Sequence.stop == 0)

        sequences = sequences.order_by(Sequence.v_length)

        try:
            tree = self.get_tree(clone_inst, sequences)
            if not tree:
                logger.warning('No sequences to make tree for clone {}'.format(
                    clone_id))
                return
        except Exception as e:
            logger.error('Error running clone {}: {}'.format(clone_id, e))
            return

        for node_id, node in enumerate(tree.traverse()):
            node.add_feature('node_id', node_id)
        final = {
            'info': {
                'min_mut_copies': self.min_mut_copies,
                'min_mut_samples': self.min_mut_samples,
                'min_seq_copies': self.min_seq_copies,
                'min_seq_samples': self.min_seq_samples,
                'exclude_stops': self.exclude_stops,
                'full_seq': self.full_seq,
            },
            'tree': tree_as_dict(tree)
        }
        clone_inst.tree = json.dumps(final)
        self.session.add(clone_inst)
        self.session.commit()


def get_fasta_input(germline_seq, sequences, min_mut_copies, min_mut_samples,
                    limit=None):
    seqs = {}
    mut_counts = {}
    for seq in sequences:
        seqs[seq.ai] = seq.clone_sequence
        if seq.mutations_from_clone is None:
            raise Exception(
                'Mutation information not available for sequence '
                '{}. Skipping this clone tree. Was '
                'immunedb_clone_stats run?'.format(seq.seq_id)
            )

        mutations = get_mutations(
            germline_seq, seq.clone_sequence, map(
                int, json.loads(seq.mutations_from_clone).keys())
        )
        if limit:
            mutations = set([m for m in mutations if m[0] < limit])
        for mut in mutations:
            if mut not in mut_counts:
                mut_counts[mut] = {'count': 0, 'samples': set([])}
            mut_counts[mut]['count'] += seq.copy_number
            mut_counts[mut]['samples'].add(seq.sample_id)

    removed_muts = set([])
    for mut, cnts in mut_counts.items():
        if (cnts['count'] < min_mut_copies or
                len(cnts['samples']) < min_mut_samples):
            removed_muts.add(mut)

    for seq_id, seq in seqs.items():
        seqs[seq_id] = remove_muts(seq, removed_muts, germline_seq)

    in_data = '>germline\n{}\n'.format(germline_seq[:limit])
    for seq_id, seq in seqs.items():
        in_data += '>{}\n{}\n'.format(seq_id, seq[:limit])
    return in_data, removed_muts


def add_tree_metadata(session, newick, germline_seq, removed_muts, limit=None):
    tree = ete3.Tree(newick)
    for node in tree.traverse():
        if node.name not in ('NoName', 'germline', ''):
            seq = session.query(Sequence).filter(
                Sequence.ai == node.name
            ).first()
            seq_ids = {}
            for collapsed_seq in get_seqs_collapsed_to(session, seq):
                seq_ids[collapsed_seq.seq_id] = {
                    'ai': collapsed_seq.ai,
                    'copy_number': collapsed_seq.copy_number,
                    'sample_name': collapsed_seq.sample.name,
                    'sample_id': collapsed_seq.sample.id,
                    'metadata': collapsed_seq.sample.metadata_dict
                }

            node.name = seq.seq_id
            node.add_feature('seq_ids', seq_ids)
            node.add_feature('copy_number', sum(
                [s['copy_number'] for s in seq_ids.values()]
            ))
            modified_seq = remove_muts(seq.sequence,
                                       removed_muts, germline_seq)
            muts = get_mutations(
                germline_seq, modified_seq,
                map(int, json.loads(seq.mutations_from_clone).keys())
            )
            if limit:
                muts = set([m for m in muts if m[0] < limit])
            node.add_feature('mutations', muts)
        else:
            node = instantiate_node(node)

    return tree


def get_seqs_collapsed_to(session, seq):
    sample_level_seqs = session.query(SequenceCollapse.seq_ai).filter(
        SequenceCollapse.collapse_to_subject_seq_ai == seq.ai
    )
    for sample_seq in sample_level_seqs:
        yield session.query(
            Sequence
        ).filter(Sequence.ai == sample_seq.seq_ai).one()


def remove_muts(seq, removes, germline_seq):
    for mut in removes:
        loc, _, to = mut
        loc -= 1
        if seq[loc] == to:
            seq = seq[:loc] + germline_seq[loc] + seq[loc + 1:]
    return seq


def get_mutations(s1, s2, positions):
    return set([(i + 1, s1[i], s2[i]) for i in positions if s1[i] != s2[i]])


def instantiate_node(node):
    node.add_feature('seq_ids', {})
    node.add_feature('copy_number', 0)
    node.add_feature('sequence', None)
    node.add_feature('mutations', set([]))

    return node


def get_nested(seqs, key):
    return sorted(set(
        [s['metadata'][key] for s in seqs.values() if key in s['metadata']]
    ))


def tree_as_dict(tree, root=True):
    all_meta = set(
        [k for seq in tree.seq_ids.values() for k in seq['metadata'].keys()]
    )
    node = {
        'data': {
            'seq_ids': tree.seq_ids,
            'node_id': tree.node_id,
            'copy_number': tree.copy_number,
            'metadata': {
                k: get_nested(tree.seq_ids, k) for k in all_meta
            },
            'mutations': [{
                'pos': mut[0],
                'from': mut[1],
                'to': mut[2],
            } for mut in tree.mutations]
        },
        'children': []
    }
    for i, child in enumerate(tree.children):
        node['children'].append(tree_as_dict(child, root=False))

    if not root or (len(tree.mutations) == 0 and len(tree.seq_ids) == 0):
        return node

    return {
        'data': {
            'seq_ids': {},
            'copy_number': 0,
            'metadata': {},
            'mutations': [],
        },
        'children': [node]
    }


def cut_tree(tree, max_muts, d=0):
    max_muts -= len(tree.mutations)
    if tree.is_leaf() or max_muts <= 0:
        return [tree]
    subtrees = []
    for node in tree.children:
        subtrees.extend(cut_tree(node, max_muts, d+1))
    return subtrees


def get_seq_pks(tree):
    seq_pks = set([
        (s['sample_id'], s['ai']) for s in tree.seq_ids.values()
    ])
    for child in tree.children:
        seq_pks.update(get_seq_pks(child))
    return seq_pks



import collections
import csv
from functools import partial
import math
import os
import json
import shlex
import subprocess

import immunedb.common.config as config
from immunedb.common.models import (Clone, CloneStats, SelectionPressure,
                                    Sequence, SequenceCollapse)
import immunedb.common.modification_log as mod_log
import immunedb.util.concurrent as concurrent
from immunedb.util.log import logger

TEST_FOCUSED = 1
TEST_LOCAL = 2
SPECIES_HUMAN = 1
SPECIES_MOUSE = 2
SUB_UNIFORM = 0
SUB_SMITH = 1
MUT_UNIFORM = 0
MUT_SHAPIRO = 1
CONSTANT_BOUNDARIES = [1, 26, 38, 55, 65, 104]

SEQ_CLONAL = 1
FIX_INDELS = 1


def get_selection(session, clone_id, script_path, samples=None,
                  min_mut_count=1,
                  max_mut_count=None,
                  temp_dir='/tmp',
                  test_type=TEST_FOCUSED,
                  species=SPECIES_HUMAN,
                  sub_model=SUB_UNIFORM,
                  mut_model=MUT_UNIFORM):
    clone = session.query(Clone).filter(Clone.id == clone_id).first()
    last_region = CONSTANT_BOUNDARIES[-1] + clone.cdr3_num_nts // 3
    boundaries = '{}:{}'.format(':'.join(map(str, CONSTANT_BOUNDARIES)),
                                last_region)
    if samples is not None:
        unique_id = '_{}_{}'.format(clone_id, '_'.join(map(str, samples)))
    else:
        unique_id = '_{}_{}'.format(clone_id, '_ALL')
    input_path = os.path.join(temp_dir, 'clone{}.fasta'.format(unique_id))
    out_path = os.path.join(temp_dir, 'output{}'.format(unique_id))
    read_path = os.path.join(temp_dir, 'output{}{}.txt'.format(unique_id,
                             clone.id))

    _make_input_file(session, input_path, clone, samples, min_mut_count,
                     max_mut_count)
    cmd = 'Rscript {} {} {} {} {} {} {} {} {} {} {}'.format(
        script_path, test_type, species,
        sub_model, mut_model, SEQ_CLONAL,
        FIX_INDELS, boundaries, input_path, out_path,
        clone.id)
    proc = subprocess.Popen(shlex.split(cmd),
                            cwd=os.path.dirname(script_path),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    proc.communicate()

    with open(read_path) as fh:
        output = _parse_output(session, clone, fh)

    os.unlink(input_path)
    os.unlink(read_path)
    os.unlink(os.path.join(
        temp_dir, 'output{}{}.RData'.format(unique_id, clone.id)))

    return output


def _make_input_file(session, input_path, clone, samples, min_mut_count,
                     max_mut_count):
    with open(input_path, 'w+') as fh:
        fh.write('>>>CLONE\n')
        fh.write('>>germline\n')
        fh.write('{}\n'.format(clone.consensus_germline))

        seqs = session.query(
            Sequence.sequence,
            Sequence.mutations_from_clone
        ).join(SequenceCollapse).filter(
            Sequence.clone == clone
        )
        if samples is None:
            seqs = seqs.filter(SequenceCollapse.copy_number_in_subject > 0)
        else:
            seqs = seqs.filter(Sequence.sample_id.in_(samples),
                               Sequence.copy_number > 0)

        if min_mut_count > 1 or max_mut_count is not None:
            removes = collections.Counter()
            seqs = seqs.all()
            # Iterate over each sequence and increment the count for each
            # mutation in the counter
            for seq in seqs:
                removes.update({
                    (i, seq.sequence[i]): 1 for i in
                    map(int, json.loads(seq.mutations_from_clone).keys())
                })

            # Filter out the mutations
            removes = [
                mut for mut, cnt in removes.items()
                if cnt < min_mut_count or cnt > (max_mut_count or -1)
            ]

            # Remove the remaining mutations
            updated_seqs = []
            for seq in seqs:
                ns = list(seq.sequence)
                for pos, to_nt in removes:
                    if ns[pos] == to_nt:
                        ns[pos] = clone.consensus_germline[pos]
                updated_seqs.append(''.join(ns))
        else:
            updated_seqs = [s.sequence for s in seqs]

        for i, seq in enumerate(updated_seqs):
            fh.write('>{}\n{}\n'.format(i, seq))


def _parse_output(session, clone, fh):
    reader = csv.DictReader(fh, delimiter='\t')
    for row in reader:
        if row['Type'] == 'Sequence':
            del row['Type']
            del row['ID']
            row = {k: v.strip() for k, v in row.items()}
            row = {
                k: v.strip() if v == 'NA' else float(v.strip()) for k, v in
                row.items()
            }
            return row


def na_to_null(v):
    return v if v != 'NA' else None


class SelectionPressureWorker(concurrent.Worker):
    """A worker class for calculating selection pressure.  This worker will
    accept one clone at a time for parallelization.

    :param Session session: The database session

    """
    def __init__(self, session, baseline_path, baseline_temp, thresholds):
        self._session = session
        self._baseline_path = baseline_path
        self._baseline_temp = baseline_temp
        self._thresholds = thresholds

    def do_task(self, clone_id):
        """Starts the task of calculation of clonal selection pressure.

        :param int args: The clone_id

        """

        if self._session.query(SelectionPressure).filter(
                SelectionPressure.clone_id == clone_id).count() > 0:
            return
        self.info('Clone {}'.format(clone_id))
        sample_ids = [c.sample_id for c in self._session.query(
                CloneStats.sample_id
            ).filter(
                CloneStats.clone_id == clone_id
            )
        ]

        for sample_id in sample_ids:
            self._process_sample(clone_id, sample_id)
        self._session.commit()

    def _process_sample(self, clone_id, sample_id):
        """Processes selection pressure for one sample (or the aggregate of all
        samples).  If ``sample_id`` is None the pressure for all sequences in
        the clone is calculated.  If ``single`` is specified, the clone only
        occurs in one sample and the entry with ``sample_id=None`` should be
        the same as for the one sample.

        :param int clone_id: The ID of the clone
        :param int sample_id: The ID of a sample in which the clone exists

        """

        total_seqs = int(self._session.query(CloneStats.unique_cnt).filter(
            CloneStats.clone_id == clone_id,
            CloneStats.sample_id == sample_id
        ).one().unique_cnt)

        base_call = partial(
            get_selection,
            session=self._session,
            clone_id=clone_id,
            script_path=self._baseline_path,
            samples=[sample_id] if sample_id is not None else None,
            temp_dir=self._baseline_temp
        )

        for threshold in self._thresholds:
            if threshold.endswith('E'):
                min_seqs = max_seqs = int(threshold[:-1])
            else:
                min_seqs = int(
                    math.ceil(int(threshold[:-1]) / 100.0 * total_seqs)
                    if '%' in threshold
                    else threshold
                )
                max_seqs = None

            pressure = base_call(min_mut_count=min_seqs,
                                 max_mut_count=max_seqs)
            pressure = {k: na_to_null(v) for k, v in pressure.items()}
            self._session.add(SelectionPressure(
                clone_id=clone_id,
                sample_id=sample_id,
                threshold=threshold,

                expected_fwr_s=pressure['Expected_FWR_S'],
                expected_cdr_s=pressure['Expected_CDR_S'],
                expected_fwr_r=pressure['Expected_FWR_R'],
                expected_cdr_r=pressure['Expected_CDR_R'],

                observed_fwr_s=pressure['Observed_FWR_S'],
                observed_cdr_s=pressure['Observed_CDR_S'],
                observed_fwr_r=pressure['Observed_FWR_R'],
                observed_cdr_r=pressure['Observed_CDR_R'],

                sigma_fwr=pressure['Focused_Sigma_FWR'],
                sigma_cdr=pressure['Focused_Sigma_CDR'],

                sigma_fwr_cilower=pressure['Focused_CIlower_FWR'],
                sigma_fwr_ciupper=pressure['Focused_CIupper_FWR'],
                sigma_cdr_cilower=pressure['Focused_CIlower_CDR'],
                sigma_cdr_ciupper=pressure['Focused_CIupper_CDR'],

                sigma_p_fwr=pressure['Focused_P_FWR'],
                sigma_p_cdr=pressure['Focused_P_CDR'],
            ))

    def cleanup(self):
        self._session.commit()
        self._session.close()


def run_selection_pressure(session, args):
    mod_log.make_mod('clone_pressure', session=session, commit=True,
                     info=vars(args))

    if args.clone_ids is not None:
        clones = args.clone_ids
    elif args.subject_ids is not None:
        clones = [c.id for c in session.query(Clone.id).filter(
            Clone.subject_id.in_(args.subject_ids))]
    else:
        clones = [c.id for c in session.query(Clone.id)]
    clones.sort()

    tasks = concurrent.TaskQueue()
    logger.info('Creating task queue to calculate selection pressure for {} '
                'clones.'.format(len(clones)))

    if args.regen:
        logger.info('Deleting old selection pressure')
        for clone in clones:
            session.query(SelectionPressure).filter(
                SelectionPressure.clone_id == clone).delete()
        session.commit()

    for cid in clones:
        tasks.add_task(cid)

    for i in range(0, args.nproc):
        session = config.init_db(args.db_config)
        tasks.add_worker(SelectionPressureWorker(session, args.baseline_path,
                                                 args.temp, args.thresholds))

    tasks.start()



import csv
import os
import re
import time
import itertools

import dnautils

from immunedb.identification import (add_noresults_for_vdj, add_sequences,
                                     AlignmentException)
from immunedb.identification.genes import GeneName, JGermlines
from immunedb.identification.vdj_sequence import VDJAlignment, VDJSequence
from immunedb.identification.metadata import (parse_metadata, REQUIRED_FIELDS,
                                              MetadataException)
from immunedb.common.models import (CDR3_OFFSET, Sample, SampleMetadata, Study,
                                    Subject)
import immunedb.util.concurrent as concurrent
import immunedb.util.funcs as funcs
from immunedb.util.log import logger
from Bio import SeqIO

from immunedb.identification.identify import IdentificationProps


class CachedTies(dict):
    def __init__(self, gene, *args, **kw):
        super(CachedTies, self).__init__(*args, **kw)
        self.gene = gene
        self.ties = {}

    def get_ties(self, genes):
        length = len(self[genes[0]])
        key = tuple(sorted(genes))
        if key in self.ties:
            return self.ties[key]
        step = 1 if self.gene == 'v' else -1
        seqs = [self[v].ljust(length, 'N')[::step] for v in genes]
        cons = funcs.consensus(seqs)[::step]
        return self.ties.setdefault(key, cons)


def raw_germlines(fn, gene):
    assert gene in ('v', 'j')

    with open(fn) as fh:
        return CachedTies(gene, {
            r.description: str(r.seq).upper().replace('.', '-')
            for r in SeqIO.parse(fh, 'fasta')
        })


def create_sample(session, metadata):
    study, new = funcs.get_or_create(
        session, Study, name=metadata['study_name'])

    if new:
        logger.info('Created new study "{}"'.format(study.name))
        session.commit()

    sample, new = funcs.get_or_create(
        session, Sample, name=metadata['sample_name'], study=study)
    if new:
        logger.info('Created new sample "{}"'.format(sample.name))
        for key, value in metadata.items():
            if key not in REQUIRED_FIELDS:
                session.add(SampleMetadata(
                    sample=sample,
                    key=key,
                    value=value
                ))

        subject, new = funcs.get_or_create(
            session, Subject, study=study,
            identifier=metadata['subject'])
        sample.subject = subject
        session.commit()
    else:
        logger.error(
            'Sample "{}" already exists'.format(metadata['sample_name']))
        return
    return sample


def collapse_duplicates(bucket):
    use_heuristic = len(bucket) > 10000
    bucket = sorted(bucket, key=lambda s: s.sequence.copy_number, reverse=True)

    if use_heuristic:
        uniques = {}
        for alignment in bucket:
            if alignment.cdr3 not in uniques:
                uniques[alignment.cdr3] = alignment
            else:
                uniques[alignment.cdr3].sequence.copy_number += (
                    alignment.sequence.copy_number)

        logger.info(
            'Bucket {v_gene} {j_gene} {cdr3_num_nts} had {cnt} '
            'sequences.  Used heuristic to reduce to {new_cnt}.'.format(
                v_gene=[g.name for g in bucket[0].v_gene],
                j_gene=[g.name for g in bucket[0].j_gene],
                cdr3_num_nts=bucket[0].cdr3_num_nts,
                cnt=len(bucket),
                new_cnt=len(uniques))
        )
        bucket = sorted(
            uniques.values(),
            key=lambda s: s.sequence.copy_number,
            reverse=True
        )

    return collapse_duplicate_alignments(bucket)


def collapse_duplicate_alignments(bucket):
    uniques = []
    while bucket:
        alignment = bucket.pop()
        for i, other_alignment in enumerate(bucket):
            if (len(alignment.sequence.sequence) !=
                    len(other_alignment.sequence.sequence)):
                logger.warning('Sequence lengths differ {} {}'.format(
                    alignment.sequence.seq_id,
                    other_alignment.sequence.seq_id)
                )
                continue
            if dnautils.equal(alignment.sequence.sequence,
                              other_alignment.sequence.sequence):
                alignment.sequence.copy_number += (
                    other_alignment.sequence.copy_number
                )
                bucket.pop(i)
        uniques.append(alignment)
    return uniques


def process_line(line, alignment_func, props, v_germlines, j_germlines):
    try:
        alignment = alignment_func(line, v_germlines, j_germlines)
        if props.trim_to:
            alignment.trim_to(props.trim_to)
        props.validate(alignment)
        return {
            'status': 'success',
            'alignment': alignment
        }
    except AlignmentException as e:
        if len(e.args) == 1:
            vdj, msg = alignment.sequence, str(e)
        else:
            vdj, msg = e.args

        return {
            'status': 'noresult',
            'vdj': vdj,
            'reason': msg
        }


def aggregate_results(results, session, sample):
    alignments = {}
    success = [r for r in results if r['status'] == 'success']
    noresults = [r for r in results if r['status'] == 'noresult']
    logger.info('{} total sequences ({} alignments, {} noresults)'.format(
        len(results), len(success), len(noresults)))

    for result in success:
        alignment = result['alignment']
        key = (
            funcs.format_ties(alignment.v_gene),
            funcs.format_ties(alignment.j_gene),
            alignment.cdr3_num_nts,
            tuple(alignment.insertions),
            tuple(alignment.deletions)
        )
        alignments.setdefault(key, []).append(alignment)

    copies = 0
    for i, result in enumerate(noresults):
        logger.debug('noresult {} / {}'.format(i, len(noresults)))
        orig_id = result['vdj'].seq_id
        copies += result['vdj'].copy_number
        for i in range(result['vdj'].copy_number):
            result['vdj'].seq_id = '{}_{}'.format(orig_id, i)
            add_noresults_for_vdj(session, result['vdj'], sample,
                                  result['reason'])
        if copies % 1000 == 0:
            session.commit()

    session.commit()
    return alignments


def add_results(uniques, sample, session):
    metrics = {'muts': [], 'lens': []}
    for unique in funcs.periodic_commit(session, itertools.chain(*uniques),
                                        interval=1000):
        try:
            add_sequences(session, [unique], sample)
            metrics['lens'].append(unique.v_length)
            metrics['muts'].append(unique.v_mutation_fraction)
        except AlignmentException as e:
            add_noresults_for_vdj(session, unique.sequence, sample, str(e))

    if metrics['lens']:
        sample.v_ties_len = sum(metrics['lens']) / len(metrics['lens'])
        sample.v_ties_mutations = sum(metrics['muts']) / len(metrics['muts'])
    session.commit()


def parse_file(fh, sample, session, alignment_func, props, v_germlines,
               j_germlines, nproc, preprocess_func=None):
    start = time.time()
    reader = csv.DictReader(fh, delimiter='\t')
    if preprocess_func:
        reader = preprocess_func(reader)

    alignments = concurrent.process_data(
        reader,
        process_line,
        aggregate_results,
        nproc,
        process_args={
            'alignment_func': alignment_func,
            'props': props,
            'v_germlines': v_germlines,
            'j_germlines': j_germlines
        },
        aggregate_args={
            'session': session,
            'sample': sample
        }
    )

    logger.info('There are {} buckets to collapse'.format(len(alignments)))
    concurrent.process_data(
        alignments.values(),
        collapse_duplicates,
        add_results,
        nproc,
        aggregate_args={
            'session': session,
            'sample': sample
        }
    )

    logger.info('Completed sample {} in {}m'.format(
        sample.name, round((time.time() - start) / 60., 1)))


def add_imgt_gaps(germline, sequence):
    seq_gaps = [i for i, c in enumerate(sequence) if c == '.']
    added = 0
    for i, c in enumerate(germline):
        if c == '-':
            gaps_before = len([j for j in seq_gaps if j < i])
            sequence.add_gap(i + gaps_before, '.')
            added += 1
    return sequence, added


def preprocess_airr(reader):
    logger.info('Collapsing identical sequences')
    seen = {}
    for l in reader:
        copies = re.search(r'DUPCOUNT=(\d+)', l['sequence_id'])
        copies = int(copies.group(1)) if copies else 1
        l['sequence_id'] = l['sequence_id'].split(
            '|DUPCOUNT')[0].replace('reversed|', '')
        if l['sequence_alignment'] in seen:
            seen[l['sequence_alignment']]['copy_number'] += copies
        else:
            l['copy_number'] = copies
            seen[l['sequence_alignment']] = l
    logger.info('Sample has {} total copies, {} unique'.format(
        sum([s['copy_number'] for s in seen.values()]),
        len(seen)
    ))
    return sorted(seen.values(), key=lambda s: s['sequence_id'])


def parse_airr(line, v_germlines, j_germlines):
    seq = VDJSequence(
        seq_id=line['sequence_id'].replace('reversed|', ''),
        sequence=line['sequence_alignment'],
        copy_number=line['copy_number']
    )
    if not all([line['sequence_id'], line['v_call'], line['j_call'],
                line['junction_aa']]):
        raise AlignmentException(seq, 'Missing v_gene, j_gene, or junction_aa')

    seq.pad(int(line['v_germline_start']) - 1)
    try:
        v_germ_seq = v_germlines.get_ties(line['v_call'].split(','))
    except KeyError:
        raise AlignmentException(
            seq,
            'V-gene {} not in germline database'.format(line['v_call'])
        )

    aligned_germ = ''.join([
        v_germ_seq.replace('-', '')[:int(line['v_germline_start']) - 1],
        line['germline_alignment']
    ])
    # Append the missing portion, if any, of the J to the germline
    j_germ_seq = j_germlines.get_ties(line['j_call'].split(','))
    append_j = len(j_germ_seq) - int(line['j_germline_end'])

    if append_j > JGermlines.defaults['upstream_of_cdr3']:
        raise AlignmentException(seq, 'No sequencing data past the CDR3')
    if append_j > 0:
        aligned_germ += j_germ_seq[-append_j:]
        seq.pad_right(append_j)

    aligned_seq, gaps_added = add_imgt_gaps(v_germ_seq, seq)
    aligned_germ = add_imgt_gaps(
        v_germ_seq, VDJSequence('', aligned_germ)
    )[0].sequence
    cdr3_start = int(line['cdr3_start']) - int(line['v_sequence_start'])
    # Push the start of the CDR3 based on number of IMGT gaps added.  Then add
    # 3 because IgBLAST's CDR3 excludes the preserved Cysteine
    cdr3_start += gaps_added - 3
    cdr3_start += aligned_seq.sequence[:cdr3_start].count('-')
    cdr3_start += int(line['v_germline_start']) - 1
    cdr3_end = cdr3_start + len(line['cdr3']) + 6
    # If there is an insertion in the CDR3 but not junction, increase CDR3
    # length
    junction_insertions = aligned_germ[cdr3_end - 3:cdr3_end].count('-')
    cdr3_end += junction_insertions

    junction_insertions = aligned_germ[cdr3_start:cdr3_start + 3].count('-')
    cdr3_start -= junction_insertions

    cdr3_seq = aligned_seq.sequence[cdr3_start:cdr3_end]
    cdr3_deletions = cdr3_seq.count('-')

    germline_cdr3 = aligned_germ[cdr3_start:cdr3_end]

    aligned_germ = ''.join([
        aligned_germ[:cdr3_start],
        '.' * (cdr3_end - cdr3_start),
        aligned_germ[cdr3_end + cdr3_deletions:].ljust(
            JGermlines.defaults['upstream_of_cdr3'], 'N'
        )
    ])
    gap = re.match('[.]+', aligned_germ)
    if gap:
        aligned_germ = ('N' * gap.end()) + aligned_germ[gap.end():]

    aligned_seq = ''.join([
        aligned_seq.sequence[:cdr3_start],
        cdr3_seq.replace('-', ''),
        aligned_seq.sequence[cdr3_end:].ljust(
            JGermlines.defaults['upstream_of_cdr3'], 'N'
        )
    ])
    gap = re.match('[.]+', aligned_seq)
    if gap:
        aligned_seq = ('N' * gap.end()) + aligned_seq[gap.end():]

    total_insertions = line['v_germline_alignment'].count('-')
    correct_cdr3_start = CDR3_OFFSET + total_insertions
    if cdr3_start != correct_cdr3_start:
        raise AlignmentException(
            seq, 'CDR3 starts at {} instead of {} ({} insertions)'.format(
                cdr3_start, correct_cdr3_start, total_insertions))

    if len(aligned_germ) != len(aligned_seq):
        raise AlignmentException(
            seq, 'Germline length {} != sequence length {}'.format(
                len(aligned_germ), len(aligned_seq)))

    alignment = funcs.ClassProxy(VDJAlignment(
        VDJSequence(
            line['sequence_id'],
            aligned_seq.replace('.', '-'),
            rev_comp=line['rev_comp'] == 'T',
            copy_number=line['copy_number'])))
    alignment.germline = aligned_germ.replace('.', '-')
    alignment.v_gene = set([GeneName(c) for c in line['v_call'].split(',')])
    alignment.j_gene = set([GeneName(c) for c in line['j_call'].split(',')])
    alignment.cdr3_start = cdr3_start
    alignment.cdr3_num_nts = len(cdr3_seq)
    alignment.locally_aligned = True
    alignment.germline_cdr3 = germline_cdr3
    alignment.seq_offset = int(line['v_germline_start']) - 1
    alignment.v_length = int(line['v_alignment_end'])
    alignment.j_length = (int(line['j_alignment_end']) -
                          int(line['j_alignment_start']))
    alignment.v_mutation_fraction = (100 - float(line['v_identity'])) / 100
    # Skipping the germline_cdr3 field and instead populating its dependencies
    # via the proxy
    alignment.j_match = float(line['j_identity']) * alignment.j_length / 100
    alignment.post_cdr3_length = len(alignment.sequence.sequence) - cdr3_end
    alignment.insertions = funcs.gap_positions(aligned_germ)
    alignment.deletions = funcs.gap_positions(aligned_seq)

    return alignment


def import_alignments(session, args):
    parse_funcs = {
        'airr': (parse_airr, preprocess_airr),
    }

    meta_fn = args.metadata if args.metadata else os.path.join(
        args.sample_dir, 'metadata.tsv')
    if not os.path.isfile(meta_fn):
        logger.error('Metadata file not found.')
        return
    with open(meta_fn, 'rU') as fh:
        try:
            metadata = parse_metadata(session, fh, args.warn_existing,
                                      args.warn_missing, args.sample_dir)
        except MetadataException as ex:
            logger.error(ex)
            return

    props = IdentificationProps(**args.__dict__)
    v_germlines = raw_germlines(args.v_germlines, 'v')
    j_germlines = raw_germlines(args.j_germlines, 'j')

    for sample_name in sorted(metadata.keys()):
        sample = create_sample(session, metadata[sample_name])
        if sample:
            path = os.path.join(
                args.sample_dir, metadata[sample_name]['file_name'])
            with open(path) as fh:
                parse_file(fh, sample, session, parse_funcs[args.format][0],
                           props, v_germlines, j_germlines,
                           args.nproc,
                           preprocess_func=parse_funcs[args.format][1])



from .sequences import *  # noqa: F401, F403
from .genbank import *  # noqa: F401, F403
from .selection import *  # noqa: F401, F403
from .samples import *  # noqa: F401, F403
from .clones.listing import *  # noqa: F401, F403
from .clones.overlap import *  # noqa: F401, F403



import getpass
import json
import pymysql
import os
import random
import re
import shlex
import string
import subprocess

import immunedb.common.config as config
from immunedb.util.log import logger


def _yn_prompt(prompt):
    while True:
        print('{} [Y/N]'.format(prompt), end=' ')
        v = input().lower()
        if v in ('y', 'n'):
            return v == 'y'


def _connect(host, user, admin_pass=''):
    return pymysql.connect(host=host, user=user, password=admin_pass,
                           cursorclass=pymysql.cursors.DictCursor)


def _get_root_connection(host, user, admin_pass=None):
    if admin_pass is None:
        try:
            return _connect(host, user)
        except Exception:
            logger.info('Failed connection with empty root password')
            admin_pass = getpass.getpass('MySQL password for ({}):'.format(
                user))
    return _connect(host, user, admin_pass)


def _create_user_if_not_exists(conn, host, user, password):
    with conn.cursor() as cursor:
        cursor.execute('SELECT host, user, authentication_string from '
                       'mysql.user WHERE user=%s and host=%s', (user, host))
        existing = cursor.fetchone()
        if existing is None:
            cursor.execute('CREATE USER \'{}\'@\'%\' IDENTIFIED BY '
                           '\'{}\''.format(user, password))
            return None
        return existing['authentication_string']


def _get_user_pass(conn, host, user, existing_password):
    with conn.cursor() as cursor:
        while True:
            db_pass = getpass.getpass()
            cursor.execute('SELECT PASSWORD(%s) as password', db_pass)
            if cursor.fetchone()['password'] != existing_password:
                logger.error('Password does not match.')
            else:
                logger.info('Correct password')
                return db_pass


def create(main_parser, args):
    if re.search(r'[^A-Za-z0-9_-]', args.db_name) is not None:
        main_parser.error('Database name must only contain letters, numbers, '
                          'dashes and underscores.')

    try:
        conn = _get_root_connection(args.db_host, args.admin_user,
                                    args.admin_pass)

        db_user = args.db_user or args.db_name
        if args.db_pass:
            db_pass = args.db_pass
        else:
            db_pass = ''.join(
                random.choice(string.ascii_uppercase + string.ascii_lowercase +
                              string.digits) for _ in range(10))

        with conn.cursor() as cursor:
            logger.info('Creating user "{}"'.format(db_user))
            existing_password = _create_user_if_not_exists(conn, '%', db_user,
                                                           db_pass)
            if existing_password is not None:
                if not args.db_pass:
                    logger.warning(
                        'User {} already exists.  To generate the '
                        'configuration file, you must enter it\'s '
                        'password.'.format(db_user)
                    )
                    db_pass = _get_user_pass(conn, args.db_host, db_user,
                                             existing_password)
                else:
                    db_pass = args.db_pass

            logger.info('Creating database "{}"'.format(args.db_name))
            cursor.execute('CREATE DATABASE {}'.format(args.db_name))

            cursor.execute(
                'GRANT ALL PRIVILEGES ON {}.* TO \'{}\'@\'%\''.format(
                    args.db_name, db_user))

        config_path = os.path.join(args.config_dir, '{}.json'.format(
            args.db_name))
        logger.info('Creating config at {}'.format(config_path))
        with open(config_path, 'w+') as fh:
            json.dump({
                'host': args.db_host,
                'database': args.db_name,
                'username': db_user,
                'password': db_pass
            }, fh, sort_keys=True, indent=4, separators=(',', ': '))

        logger.info('Initializing tables')
        config.init_db(config_path)
        logger.info('Success!')
        return True
    except Exception as e:
        logger.error(e)
        return False


def delete(main_parser, args):
    try:
        with open(args.db_config) as fh:
            db_config = json.load(fh)
        conn = _get_root_connection(db_config['host'], args.admin_user,
                                    args.admin_pass)
        with conn.cursor() as cursor:
            logger.info('Deleting database {}'.format(db_config['database']))
            cursor.execute('DROP DATABASE `{}`'.format(db_config['database']))
            if args.delete_user:
                logger.info('Deleting user {}'.format(db_config['username']))
                cursor.execute('DROP USER `{}`'.format(db_config['username']))
        return True
    except Exception as e:
        logger.error(e)
        return False


def backup(main_parser, args):
    with open(args.db_config) as fh:
        db_config = json.load(fh)
    with open(args.backup_path, 'w+') as fh:
        cmd = shlex.split(
            'mysqldump -h {} -u {} -p{} {}'.format(db_config['host'],
                                                   db_config['username'],
                                                   db_config['password'],
                                                   db_config['database']))
        proc = subprocess.Popen(cmd, stdout=fh)
        proc.communicate()

    return True


def restore(main_parser, args):
    with open(args.db_config) as fh:
        db_config = json.load(fh)
    with open(args.backup_path, 'r') as fh:
        cmd = shlex.split(
            'mysql -h {} -u {} -p{} {}'.format(db_config['host'],
                                               db_config['username'],
                                               db_config['password'],
                                               db_config['database']))
        proc = subprocess.Popen(cmd, stdin=fh, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        if stderr:
            logger.warning(stderr)
    return True



import shlex
from subprocess import Popen, PIPE

import ete3

import immunedb.common.config as config
from immunedb.common.models import Clone
import immunedb.common.modification_log as mod_log
from immunedb.trees import instantiate_node, LineageWorker
import immunedb.util.concurrent as concurrent
from immunedb.util.log import logger


def get_newick(fasta_input):
    proc = Popen(shlex.split('clearcut --alignment -q --DNA -N -r'),
                 stdin=PIPE, stdout=PIPE, stderr=PIPE, encoding='utf8')
    return proc.communicate(input=fasta_input)[0]


def minimize_tree(tree):
    tree.set_outgroup('germline')
    tree.search_nodes(name='germline')[0].delete()
    first = True
    while True:
        push_common_mutations_up(tree, first)
        remove_parent_mutations(tree)
        rem_null = remove_null_nodes(tree)
        moved = check_supersets(tree)
        if not rem_null and not moved and not are_null_nodes(tree):
            break
        first = False
    return tree


def push_common_mutations_up(tree, first):
    if tree.is_leaf():
        return tree.mutations

    common_muts = None
    for child in tree.children:
        child_muts = push_common_mutations_up(child, first)
        if common_muts is None:
            common_muts = child_muts
        else:
            common_muts = common_muts.intersection(child_muts)

    if len(tree.seq_ids) == 0:
        if first:
            tree.mutations = common_muts or tree.mutations
        else:
            tree.mutations = common_muts.union(tree.mutations)

    return tree.mutations


def remove_parent_mutations(tree):
    for node in tree.traverse(strategy='postorder'):
        if node.up is not None and node.up.name != 'germline':
            node.mutations = node.mutations.difference(node.up.mutations)


def remove_null_nodes(tree):
    removed = False
    for node in tree.traverse():
        if node.up is not None and len(node.mutations) == 0:
            node.up.seq_ids.update(node.seq_ids)
            node.up.copy_number += node.copy_number
            node.delete(prevent_nondicotomic=False)
            removed = True
    return removed


def check_supersets(tree):
    if tree.is_leaf():
        return False

    moved = False
    for c1 in tree.children:
        for c2 in tree.children:
            if c1 == c2:
                continue
            if c1.mutations.issubset(c2.mutations):
                c1.detach()
                c2.add_child(c1)
                moved = True
            elif c2.mutations.issubset(c1.mutations):
                c2.detach()
                c1.add_child(c2)
                moved = True

            overlap = c1.mutations.intersection(c2.mutations)
            if len(overlap) > 0:
                c1.detach()
                c2.detach()
                intermediate = instantiate_node(ete3.Tree(name='NoName'))
                intermediate.mutations = overlap
                intermediate.add_child(c1)
                intermediate.add_child(c2)
                tree.add_child(intermediate)
            moved = moved or check_supersets(c1)

    return moved


def are_null_nodes(tree):
    for node in tree.traverse():
        if node.up is not None and len(node.mutations) == 0:
            return True
    return False


def run_clearcut(session, args):
    if args.clone_ids is not None:
        clones = session.query(Clone.id).filter(
            Clone.id.in_(args.clone_ids))
    else:
        if args.subject_ids is not None:
            clones = session.query(Clone.id).filter(
                Clone.subject_id.in_(args.subject_ids))
        else:
            clones = session.query(Clone.id)

    if not args.force:
        clones = clones.filter(Clone.tree.is_(None))
    clones = [c.id for c in clones]
    mod_log.make_mod('clone_tree', session=session, commit=True,
                     info=vars(args))

    tasks = concurrent.TaskQueue()

    logger.info('Creating task queue for clones')
    for clone_id in clones:
        tasks.add_task(clone_id)

    for _ in range(0, args.nproc):
        session = config.init_db(args.db_config)
        tasks.add_worker(LineageWorker(
            session, get_newick,
            args.min_mut_copies, args.min_mut_samples,
            args.min_seq_copies,
            args.min_seq_samples,
            args.exclude_stops,
            args.full_seq,
            post_tree_hook=minimize_tree))

    tasks.start()



import itertools

from immunedb.common.models import CDR3_OFFSET, NoResult, Sequence
import immunedb.util.funcs as funcs
from immunedb.util.log import logger
import immunedb.util.lookups as lookups


class AlignmentException(Exception):
    pass


def get_noresult_from_vdj(session, vdj, sample, reason):
    return NoResult(
        seq_id=vdj.seq_id,
        sample_id=sample.id,
        sequence=vdj.orig_sequence,
        quality=vdj.orig_quality,
        reason=reason
    )


def get_seq_from_alignment(session, alignment, sample, strip_alleles=True):
    try:
        return [Sequence(
            seq_id=alignment.sequence.seq_id,
            sample_id=sample.id,

            subject_id=sample.subject.id,

            partial=alignment.partial,
            rev_comp=alignment.sequence.rev_comp,

            probable_indel_or_misalign=alignment.has_possible_indel,

            v_gene=funcs.format_ties(alignment.v_gene, strip_alleles),
            j_gene=funcs.format_ties(alignment.j_gene, strip_alleles),

            num_gaps=alignment.num_gaps,
            seq_start=alignment.seq_start,

            v_match=alignment.v_match,
            v_length=alignment.v_length,
            j_match=alignment.j_match,
            j_length=alignment.j_length,

            removed_prefix=alignment.sequence.removed_prefix_sequence,
            removed_prefix_qual=alignment.sequence.removed_prefix_quality,
            v_mutation_fraction=alignment.v_mutation_fraction,

            pre_cdr3_length=alignment.pre_cdr3_length,
            pre_cdr3_match=alignment.pre_cdr3_match,
            post_cdr3_length=alignment.post_cdr3_length,
            post_cdr3_match=alignment.post_cdr3_match,

            in_frame=alignment.in_frame,
            functional=alignment.functional,
            stop=alignment.stop,
            copy_number=alignment.sequence.copy_number,

            cdr3_nt=alignment.cdr3,
            cdr3_num_nts=len(alignment.cdr3),
            cdr3_aa=lookups.aas_from_nts(alignment.cdr3),

            sequence=str(alignment.sequence.sequence),
            quality=alignment.sequence.quality,

            locally_aligned=alignment.locally_aligned,
            insertions=alignment.insertions,
            deletions=alignment.deletions,

            germline=alignment.germline)]
    except ValueError as e:
        try:
            return [get_noresult_from_vdj(session, alignment.sequence, sample,
                                          str(e))]
        except ValueError:
            return []


def add_sequences(session, alignments, sample, strip_alleles=True,
                  error_action='discard'):
    seqs_and_noresults = funcs.flatten([
        get_seq_from_alignment(session, a, sample, strip_alleles)
        for a in alignments
    ])
    succeeded = [n for n in seqs_and_noresults if type(n) == Sequence]
    failed = [n for n in seqs_and_noresults if type(n) == NoResult]
    funcs.bulk_add(session, succeeded)
    funcs.bulk_add(session, failed)
    session.flush()


def add_noresults_for_vdj(session, vdj, sample, reason):
    try:
        session.add(get_noresult_from_vdj(session, vdj, sample, reason))
    except ValueError:
        logger.warning('Unable to add noresult')


def get_common_seq(seqs, cutoff=True, right=False):
    if right:
        seqs = [reversed(s) for s in seqs]
    if len(seqs) == 0:
        return seqs[0]
    gene = [
        nts[0] if all([n == nts[0] for n in nts]) else 'N'
        for nts in itertools.zip_longest(*seqs, fillvalue='N')
    ]
    if right:
        gene = reversed(gene)
    gene = ''.join(gene)
    if cutoff:
        return gene[:CDR3_OFFSET]
    return gene



from sqlalchemy.sql import exists

import dnautils
import immunedb.common.config as config
from immunedb.common.models import (Clone, Sample, Sequence, SequenceCollapse,
                                    Subject)
import immunedb.common.modification_log as mod_log
import immunedb.util.concurrent as concurrent

from immunedb.util.log import logger


class CollapseWorker(concurrent.Worker):
    """A worker for collapsing sequences without including positions where
    either sequences has an 'N'.
    :param Session session: The database session
    """
    def __init__(self, session):
        self._session = session
        self._tasks = 0

    def do_task(self, bucket):
        seqs = self._session.query(
            Sequence.sample_id, Sequence.ai, Sequence.seq_id,
            Sequence.sequence, Sequence.copy_number
        ).filter(
            Sequence.subject_id == bucket.subject_id,
            Sequence.v_gene == bucket.v_gene,
            Sequence.j_gene == bucket.j_gene,
            Sequence.cdr3_num_nts == bucket.cdr3_num_nts,
            Sequence._insertions == bucket._insertions,
            Sequence._deletions == bucket._deletions
        )

        to_process = sorted([{
            'sample_id': s.sample_id,
            'ai': s.ai,
            'seq_id': s.seq_id,
            'sequence': s.sequence,
            'cn': s.copy_number
        } for s in seqs], key=lambda e: -e['cn'])

        while len(to_process) > 0:
            # Get the largest sequence in the list
            larger = to_process.pop(0)
            # Iterate over all smaller sequences to find matches
            instances = 1
            samples = set([larger['sample_id']])
            for i in reversed(range(len(to_process))):
                smaller = to_process[i]
                if len(larger['sequence']) != len(smaller['sequence']):
                    self.warning('Tried to collapse sequences of different '
                                 'lengths.  AIs are {} {}'.format(
                                     larger['ai'], smaller['ai']))
                elif dnautils.equal(larger['sequence'], smaller['sequence']):
                    # Add the smaller sequence's copy number to the larger
                    larger['cn'] += smaller['cn']
                    # If the smaller sequence matches the larger, collapse it
                    # to the larger
                    self._session.add(SequenceCollapse(**{
                        'sample_id': smaller['sample_id'],
                        'seq_ai': smaller['ai'],
                        'collapse_to_subject_seq_ai': larger['ai'],
                        'collapse_to_subject_sample_id': larger['sample_id'],
                        'collapse_to_subject_seq_id': larger['seq_id'],
                        'instances_in_subject': 0,
                        'copy_number_in_subject': 0,
                        'samples_in_subject': 0,
                    }))
                    instances += 1
                    samples.add(smaller['sample_id'])
                    # Delete the smaller sequence from the list to process
                    # since it's been collapsed
                    del to_process[i]

            # Update the larger sequence's copy number and "collapse" to itself
            self._session.add(SequenceCollapse(**{
                'sample_id': larger['sample_id'],
                'seq_ai': larger['ai'],
                'collapse_to_subject_sample_id': larger['sample_id'],
                'collapse_to_subject_seq_id': larger['seq_id'],
                'collapse_to_subject_seq_ai': larger['ai'],
                'instances_in_subject': instances,
                'copy_number_in_subject': larger['cn'],
                'samples_in_subject': len(samples),
            }))

        self._session.commit()
        self._tasks += 1
        if self._tasks > 0 and self._tasks % 100 == 0:
            self.info('Collapsed {} buckets'.format(self._tasks))

    def cleanup(self):
        self.info('Committing collapsed sequences')
        self._session.commit()
        self._session.close()


def run_collapse(session, args):
    mod_log.make_mod('collapse', session=session, commit=True,
                     info=vars(args))
    subject_ids = []

    subjects = (args.subject_ids or [e.id for e in session.query(Subject.id)])
    for subject in subjects:
        if session.query(Sample).filter(
                Sample.subject_id == subject,
                ~exists().where(
                    SequenceCollapse.sample_id == Sample.id
                )).first() is None:
            logger.info('Subject {} already collapsed.  Skipping.'.format(
                subject))
        else:
            logger.info('Resetting collapse info for subject {}'.format(
                subject))
            samples = session.query(Sample).filter(
                  Sample.subject_id == subject
            )
            for sample in samples:
                session.query(SequenceCollapse).filter(
                    SequenceCollapse.sample_id == sample.id
                ).delete(synchronize_session=False)
                sample.sample_stats = []
            logger.info('Resetting clone info for subject {}'.format(subject))
            session.query(Clone).filter(Clone.subject_id == subject).delete()
            subject_ids.append(subject)
    session.commit()

    logger.info('Creating task queue to collapse {} subjects.'.format(
        len(subject_ids)))

    tasks = concurrent.TaskQueue()

    for subject_id in subject_ids:
        buckets = session.query(
            Sequence.subject_id, Sequence.v_gene, Sequence.j_gene,
            Sequence.cdr3_num_nts, Sequence._insertions, Sequence._deletions
        ).filter(
            Sequence.subject_id == subject_id
        ).group_by(
            Sequence.subject_id, Sequence.v_gene, Sequence.j_gene,
            Sequence.cdr3_num_nts, Sequence._insertions, Sequence._deletions
        )
        for bucket in buckets:
            tasks.add_task(bucket)

    logger.info('Generated {} total tasks'.format(tasks.num_tasks()))

    for i in range(0, min(tasks.num_tasks(), args.nproc)):
        tasks.add_worker(CollapseWorker(config.init_db(args.db_config)))
    tasks.start()

    session.close()



import os
import uuid
from functools import partial
import multiprocessing as mp

import logging
from immunedb.util.log import logger


class JobQueue:
    def __init__(self, temp_dir='/tmp/immunedb-jobs'):
        self.temp_dir = temp_dir
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)
        self.files = []

    def get_path(self, uid, ext=''):
        return os.path.join(self.temp_dir, uid + ext)

    def _job_wrap(self, func, uid, **kwargs):
        logger.handlers = [
            logging.FileHandler(self.get_path(uid, '.log'))
        ]
        lock_fh = open(self.get_path(uid, '.LOCK'), 'wb+')
        with open(self.get_path(uid, '.zip'), 'wb+') as fh:
            val = func(**kwargs)
            fh.write(val)
        os.remove(lock_fh.name)

    def start_job(self, func, **kwargs):
        uid = str(uuid.uuid4())
        logger.info('Starting job with UUID {}'.format(uid))

        job_func = partial(self._job_wrap, func, uid)

        mp.Process(
            target=job_func,
            kwargs=kwargs
        ).start()
        self.files.extend([
            self.get_path(uid, '.log'),
            self.get_path(uid, '.zip'),
        ])

        return uid

    def job_complete(self, uid):
        if os.path.exists(self.get_path(uid, '.LOCK')):
            return False
        return True

    def get_log(self, uid):
        try:
            return open(self.get_path(uid, '.log'), 'r', buffering=1).read()
        except FileNotFoundError:
            return None

    def cleanup(self):
        for fn in self.files:
            try:
                os.remove(fn)
            except OSError as e:
                logger.warning('Could not remove {}: {}'.format(fn, e))



import json
from functools import wraps
import os
import re
import signal
import sys

import bottle
from bottle import response, request

from immunedb.api.jobs import JobQueue
import immunedb.api.queries as queries
import immunedb.common.config as config
from immunedb.util.log import logger
import immunedb.exporting as exporting


class EnableCors(object):
    """A class to enable Cross-Origin Resource Sharing to facilitate AJAX
    requests.

    """
    name = 'enable_cors'
    api = 2

    def apply(self, fn, context):
        def _enable_cors(*args, **kwargs):
            response.headers['Access-Control-Allow-Origin'] = '*'
            response.headers['Access-Control-Allow-Methods'] = ('GET, POST')
            response.headers['Access-Control-Allow-Headers'] = (
                'Origin, '
                'Accept, Content-Type, '
                'X-Requested-With, X-CSRF-Token')

            if bottle.request.method != 'OPTIONS':
                return fn(*args, **kwargs)

        return _enable_cors


def create_response(j=None, code=200, ctype='application/json'):
    if j is None:
        bottle.response.status = code
        return
    bottle.response.content_type = ctype
    bottle.response.status = code
    if ctype == 'application/json':
        return json.dumps(j)
    return j


def get_paging():
    data = bottle.request.json or {}
    if data.get('nopage', False):
        return None
    return [
        int(data.get('page', 1)),
        int(data.get('per_page', 10))
    ]


def decode_run_length(encoding):
    if not encoding:
        return []
    ids = []
    offset = 1
    for match in re.finditer(r'(T|F)(\d+)', encoding.upper()):
        size = int(match.group(2))
        if match.group(1) == 'T':
            ids.extend(range(offset, offset + size))
        offset += size
    return ids


def create_app(db_config, allow_shutdown=False):
    app = bottle.Bottle()
    app.install(EnableCors())
    job_queue = JobQueue()
    app.job_queue = job_queue

    def with_session(f):
        @wraps(f)
        def _wrapper(*args, **kwargs):
            session = config.init_db(db_config)
            try:
                return f(session, *args, **kwargs)
            except Exception:
                raise
            finally:
                session.close()

        return _wrapper

    @app.route('/samples/list/<sample_encoding>', method=['POST', 'OPTIONS'])
    @app.route('/samples/list', method=['POST', 'OPTIONS'])
    @with_session
    def samples_list(session, sample_encoding=None):
        sample_ids = decode_run_length(sample_encoding)
        return create_response(queries.get_samples(session, sample_ids))

    @app.route('/sequences/list', method=['POST', 'OPTIONS'])
    @with_session
    def sequences_list(session):
        fields = bottle.request.json or {}
        return create_response(queries.get_sequences(
            session,
            fields.get('filters', {}),
            fields.get('order_field', None),
            fields.get('order_dir', 'asc'),
            fields.get('subject_id', None),
            get_paging()))

    @app.route('/sequence/<sample_id>/<seq_id>', method=['POST', 'OPTIONS'])
    @with_session
    def sequence(session, sample_id, seq_id):
        return create_response(queries.get_sequence(session, sample_id,
                                                    seq_id))

    @app.route('/clones/list', method=['POST', 'OPTIONS'])
    @app.route('/clones/list/<subject_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clones_list(session, subject_id=None):
        fields = bottle.request.json or {}
        return create_response(queries.get_clones(
            session,
            fields.get('filters', {}),
            fields.get('order_field', None),
            fields.get('order_dir', 'asc'),
            subject_id,
            get_paging()))

    @app.route('/clone/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone(session, clone_id):
        return create_response(queries.get_clone(session, clone_id))

    @app.route('/clone/sequences/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_sequences(session, clone_id):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_sequences(
            session, clone_id, fields.get('get_collapse', False), get_paging()
        ))

    @app.route('/clone/mutations/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_mutations(session, clone_id):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_mutations(
            session,
            clone_id,
            fields.get('type', 'percent'),
            int(fields.get('value', 0))
        ))

    @app.route('/clone/lineage/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_lineage(session, clone_id):
        return create_response(queries.get_clone_tree(session, clone_id))

    @app.route('/clone/pressure/<clone_id>', method=['POST', 'OPTIONS'])
    @with_session
    def clone_pressure(session, clone_id):
        return create_response(list(
            queries.get_selection_pressure(session, clone_id)))

    @app.route('/samples/analyze/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def analyze_samples(session, sample_encoding):
        fields = bottle.request.json or {}
        return create_response(
            queries.analyze_samples(
                session,
                decode_run_length(sample_encoding),
                fields.get('filter_type', 'unique_multiple'),
                fields.get('include_outliers', True),
                fields.get('include_partials', True),
                fields.get('percentages', False),
                fields.get('grouping', 'name'),
            )
        )

    @app.route('/samples/overlap/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def overlap(session, sample_encoding):
        fields = bottle.request.json or {}
        return create_response(queries.get_clone_overlap(
            session,
            decode_run_length(sample_encoding),
            fields.get('filter_type', 'clones_all'),
            fields.get('order_by', 'total_cnt'),
            get_paging())
        )

    @app.route('/samples/v_usage/<sample_encoding>', method=[
        'POST', 'OPTIONS'])
    @with_session
    def v_usage(session, sample_encoding):
        fields = bottle.request.json or {}
        data, x_categories, totals, prefix = queries.get_v_usage(
            session,
            decode_run_length(sample_encoding),
            fields.get('filter_type', 'unique_multiple'),
            fields.get('include_outliers', True),
            fields.get('include_partials', True),
            fields.get('grouping', 'name'),
            fields.get('by_family', False)
        )

        x_categories.sort()
        y_categories = sorted(data.keys())
        array = []
        min_v = 100
        max_v = 0
        for i, x in enumerate(x_categories):
            for j, y in enumerate(y_categories):
                usage_for_y = data[y]
                if x in usage_for_y:
                    min_v = min(min_v, usage_for_y[x])
                    max_v = max(max_v, usage_for_y[x])
                    array.append([i, j, usage_for_y[x]])
                else:
                    array.append([i, j, 0])

        return create_response({
            'x_categories': ['{}{}'.format(prefix, e) for e in x_categories],
            'y_categories': y_categories,
            'totals': totals,
            'data': array,
            'min': min_v,
            'max': max_v
        })

    @app.route('/subjects/list', method=['POST', 'OPTIONS'])
    @with_session
    def subjects_list(session):
        return create_response(queries.get_all_subjects(session, get_paging()))

    @app.route('/subject/<subject_id>', method=['POST', 'OPTIONS'])
    @with_session
    def subject(session, subject_id):
        return create_response(queries.get_subject(session, subject_id))

    @app.route('/export/job_log/<uid>', method=['GET', 'OPTIONS'])
    def job_log(uid):
        log = job_queue.get_log(uid)
        if not log:
            return create_response(code=404)

        return {
            'complete': job_queue.job_complete(uid),
            'log': '\n'.join(log.split('\n')[-25:]) or ''
        }

    @app.route('/export/job/<uid>', method=['GET', 'OPTIONS'])
    def job(uid):
        if job_queue.job_complete(uid):
            return bottle.static_file(uid + '.zip', root=job_queue.temp_dir)
        return create_response(code=400)

    @app.route('/export/sequences', method=['GET', 'OPTIONS'])
    @with_session
    def export_sequences(session):
        schema = request.query.get('format')
        if schema not in exporting.mappings:
            return create_response(code=400)

        uid = job_queue.start_job(
            exporting.write_sequences,
            session=session,
            sample_ids=decode_run_length(request.query.get('samples')),
            out_format=schema,
            clones_only=request.query.get('clones_only', False),
            min_subject_copies=request.query.get('min_subject_copies', 1),
            zipped=True
        )

        return create_response({'uid': uid})

    @app.route('/export/clones', method=['GET', 'OPTIONS'])
    @with_session
    def export_clones(session):
        schema = request.query.get('format')
        if schema not in ('vdjtools', 'immunedb'):
            return create_response(code=400)

        uid = job_queue.start_job(
            exporting.write_pooled_clones,
            session=session,
            out_format=schema,
            sample_ids=decode_run_length(request.query.get('samples')),
            pool_on=request.query.get('pool_on', 'sample').split(','),
            zipped=True,
        )

        return create_response({'uid': uid})

    @app.route('/export/overlap', method=['GET', 'OPTIONS'])
    @with_session
    def export_overlap(session):
        uid = job_queue.start_job(
            exporting.write_clone_overlap,
            session=session,
            sample_ids=decode_run_length(request.query.get('samples')),
            pool_on=request.query.get('pool_on', 'sample').split(','),
            size_metric=request.query.get('size_metric'),
            sim_func=request.query.get('sim_func'),
            agg_func=request.query.get('agg_func'),
            zipped=True,
        )

        return create_response({'uid': uid})

    @app.route('/export/samples', method=['GET', 'OPTIONS'])
    @with_session
    def export_samples(session):
        uid = job_queue.start_job(
            exporting.write_samples,
            session=session,
            sample_ids=decode_run_length(request.query.get('samples')),
            zipped=True
        )
        return create_response({'uid': uid})

    @app.route('/export/selection', method=['GET', 'OPTIONS'])
    @with_session
    def export_selection(session):
        uid = job_queue.start_job(
            exporting.write_selection,
            session=session,
            sample_ids=decode_run_length(request.query.get('samples')),
            zipped=True
        )
        return create_response({'uid': uid})

    @app.route('/shutdown', method=['GET'])
    def shutdown():
        if allow_shutdown:
            logger.warning('Shutting down from remote request')
            os.kill(os.getppid(), signal.SIGINT)
        return create_response(code=404)

    return app


def run_rest_service(args):
    app = create_app(args.db_config, args.allow_shutdown)
    app.catchall = False

    try:
        # This is a hack because gunicorn tries to read the CLI args
        sys.argv = ['']

        app.run(
            host='0.0.0.0',
            port=args.port,
            server='gunicorn',
            worker_class='eventlet',
            timeout=0,
            sendfile=False,
            workers=args.nproc)
    except (KeyboardInterrupt, SystemExit):
        app.job_queue.cleanup()



from collections import Counter
import json
import math
import re

from sqlalchemy import and_, desc, distinct
from sqlalchemy.orm.strategy_options import Load
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import false, true

from immunedb.common.models import (Clone, CloneStats, Sample, SampleStats,
                                    SelectionPressure, Sequence,
                                    SequenceCollapse, Subject)
from immunedb.common.mutations import threshold_mutations


_clone_filters = {
    'clones_all': lambda q: q,
    'clones_functional': lambda q: q.filter(Clone.functional == 1),
    'clones_nonfunctional': lambda q: q.filter(Clone.functional == 0)
}


def _fields_to_dict(fields, row):
    d = {}
    for f in fields:
        d[f] = getattr(row, f)
    return d


def _subject_to_dict(subject):
    return {
        'study': {
            'id': subject.study.id,
            'name': subject.study.name
        },
        'identifier': subject.identifier,
        'id': subject.id
    }


def _sample_to_dict(sample):
    d = _fields_to_dict(['id', 'name'], sample)
    d['subject'] = _subject_to_dict(sample.subject)
    d['metadata'] = sample.metadata_dict
    return d


def _clone_to_dict(clone):
    d = _fields_to_dict(['id', 'cdr3_nt', 'v_gene', 'j_gene', 'cdr3_aa',
                         'cdr3_num_nts', 'regions', 'insertions', 'deletions',
                         'overall_unique_cnt', 'overall_total_cnt',
                         'overall_instance_cnt',
                         'overall_unique_cnt_with_subclones',
                         'overall_instance_cnt_with_subclones',
                         'overall_total_cnt_with_subclones'], clone)
    d['subject'] = _subject_to_dict(clone.subject)
    d['germline'] = clone.consensus_germline

    return d


def _page_query(q, paging):
    if paging is None:
        return q
    page, per_page = paging
    return q.offset(max(0, (page - 1) * per_page)).limit(per_page)


def get_samples(session, sample_ids=None):
    query = session.query(
        Sample, SampleStats
    ).outerjoin(
        SampleStats,
        and_(
            SampleStats.sample_id == Sample.id,
            SampleStats.outliers == true(),
            SampleStats.full_reads == false(),
            SampleStats.filter_type == 'all'
        )
    )

    if sample_ids:
        query = query.filter(SampleStats.sample_id.in_(sample_ids))

    query = query.order_by(Sample.subject_id).options(
        Load(SampleStats).load_only(
            'sequence_cnt', 'in_frame_cnt', 'stop_cnt', 'functional_cnt',
            'no_result_cnt'
        )
    )

    clone_counts = session.query(
        SampleStats.sample_id,
        SampleStats.sequence_cnt,
        SampleStats.functional_cnt
    ).filter(
        SampleStats.outliers == true(),
        SampleStats.full_reads == false(),
        SampleStats.filter_type == 'clones_all'
    )
    clone_counts = {s.sample_id: s for s in clone_counts}

    result = []
    for sample, stats in query:
        sample_dict = _sample_to_dict(sample)
        if stats is not None:
            clones = clone_counts[sample.id]
            sample_dict['sequence_cnt'] = stats.sequence_cnt
            sample_dict['functional_cnt'] = stats.functional_cnt
            sample_dict['no_result_cnt'] = stats.no_result_cnt
            sample_dict['total_cnt'] = stats.sequence_cnt + stats.no_result_cnt
            sample_dict['clone_cnt'] = clones.sequence_cnt
            sample_dict['functional_clone_cnt'] = clones.functional_cnt
        result.append(sample_dict)

    return result


def get_clones(session, filters, order_field, order_dir, subject_limit=None,
               paging=None):
    """Gets a list of all clones"""
    res = []
    clone_q = session.query(Clone)

    if subject_limit is not None:
        clone_q = clone_q.filter(Clone.subject_id == subject_limit)

    if filters is not None:
        size_field = filters.pop('size_field', 'copies')
        if size_field == 'copies':
            size_field = Clone.overall_total_cnt
        elif size_field == 'uniques':
            size_field = Clone.overall_unique_cnt
        elif size_field == 'instances':
            size_field = Clone.overall_instance_cnt

        for key, value in filters.items():
            if value is None:
                continue
            value = str(value).strip()
            if len(value) > 0 and value is not None:
                if key == 'min_cdr3_num_nts':
                    clone_q = clone_q.filter(Clone.cdr3_num_nts >= int(value))
                elif key == 'max_cdr3_num_nts':
                    clone_q = clone_q.filter(Clone.cdr3_num_nts <= int(value))
                elif key == 'cdr3_nt':
                    clone_q = clone_q.filter(Clone.cdr3_nt.like(value))
                elif key == 'cdr3_aa':
                    clone_q = clone_q.filter(Clone.cdr3_aa.like(value))
                elif key == 'min_size':
                    clone_q = clone_q.filter(size_field >= int(value))
                elif key == 'max_size':
                    clone_q = clone_q.filter(size_field <= int(value))
                elif key == 'id':
                    clone_q = clone_q.filter(Clone.id == int(value))
                elif key == 'subject_id':
                    clone_q = clone_q.filter(Clone.subject_id == int(value))
                else:
                    clone_q = clone_q.filter(
                        getattr(Clone, key).like(value))

    if order_field:
        order_field = getattr(Clone, order_field)
        if order_dir == 'asc':
            clone_q = clone_q.order_by(order_field)
        else:
            clone_q = clone_q.order_by(desc(order_field))

    for c in _page_query(clone_q, paging):
        stats_comb = []

        query = session.query(
            CloneStats.unique_cnt, CloneStats.total_cnt, Sample
        ).join(Sample).filter(
            CloneStats.clone_id == c.id,
            ~CloneStats.sample_id.is_(None)
        ).order_by(
            desc(CloneStats.unique_cnt), desc(CloneStats.total_cnt),
            desc(CloneStats.sample_id)
        )
        for stat in query:
            stats_comb.append({
                'sample': {
                    'id': stat.Sample.id,
                    'name': stat.Sample.name
                },
                'unique_sequences': int(stat.unique_cnt),
                'total_sequences': int(stat.total_cnt)
            })
        clone_dict = _clone_to_dict(c)
        clone_dict['stats'] = stats_comb
        res.append(clone_dict)

    return res


def get_clone(session, clone_id):
    result = {}
    clone = session.query(Clone).filter(Clone.id == clone_id).first()

    stats = session.query(
        CloneStats
    ).filter(
        CloneStats.clone_id == clone.id,
    ).order_by(desc(CloneStats.unique_cnt))

    result = {
        'clone': _clone_to_dict(clone),
        'parent': _clone_to_dict(clone.parent) if clone.parent else None,
        'children': [_clone_to_dict(c) for c in clone.children],
        'samples': {
            'single': []
        }
    }

    for stat in stats:
        if stat.sample_id is None:
            result['samples']['all'] = {
                'unique': stat.unique_cnt,
                'total': stat.total_cnt
            }
        else:
            sample = _sample_to_dict(stat.sample)
            sample['unique'] = stat.unique_cnt
            sample['total'] = stat.total_cnt
            result['samples']['single'].append(sample)

    return result


def get_clone_mutations(session, clone_id, threshold_type, threshold_val,
                        sample_id=None):
    clone_stats = session.query(
        CloneStats.mutations,
        CloneStats.unique_cnt
    ).filter(
        CloneStats.clone_id == clone_id,
        CloneStats.sample_id == sample_id
    ).first()
    all_mutations = json.loads(clone_stats.mutations)
    total_seqs = clone_stats.unique_cnt

    result = {
        'positions': all_mutations['positions'],
        'regions': {},
        'total_seqs': total_seqs
    }
    if threshold_type == 'sequences':
        seq_min = threshold_val
    else:
        seq_min = int(math.ceil(threshold_val / 100.0 * total_seqs))
    result['regions'] = threshold_mutations(all_mutations, seq_min)

    return result


def get_clone_sequences(session, clone_id, get_collapse, paging):
    query = session.query(
        Sequence, SequenceCollapse.copy_number_in_subject,
        SequenceCollapse.instances_in_subject
    ).join(SequenceCollapse).filter(
        Sequence.clone_id == clone_id,
        SequenceCollapse.copy_number_in_subject > 0
    ).order_by(
        desc(SequenceCollapse.copy_number_in_subject)
    )
    query = _page_query(query, paging)

    sequences = {}
    start_ptrn = re.compile(r'[N-]*')
    for seq, copy_number_in_subject, instances_in_subject in query:
        sequences[(seq.sample.id, seq.seq_id)] = {
            'seq_id': seq.seq_id,
            'sample': {
                'id': seq.sample.id,
                'name': seq.sample.name,
            },
            'cdr3_nt': seq.cdr3_nt,
            'insertions': seq.insertions,
            'deletions': seq.deletions,
            'sequence': seq.clone_sequence,
            'read_start': start_ptrn.match(seq.sequence).span()[1] or 0,
            'copy_number_in_subject': int(copy_number_in_subject),
            'instances_in_subject': int(instances_in_subject),
            'mutations': json.loads(seq.mutations_from_clone),
            'v_extent': seq.get_v_extent(in_clone=True),
            'j_length': seq.j_length,
            'collapse_to': []
        }

    if get_collapse:
        q = session.query(
            Sequence, SequenceCollapse
        ).outerjoin(SequenceCollapse).filter(
            Sequence.clone_id == clone_id,
            SequenceCollapse.copy_number_in_subject == 0
        ).order_by(desc(Sequence.copy_number))

        for seq, collapse in q:
            key = (collapse.collapse_to_subject_sample_id,
                   collapse.collapse_to_subject_seq_id)
            if key in sequences:
                sequences[key]['collapse_to'].append({
                    'sample_id': seq.sample.id,
                    'sample_name': seq.sample.name,
                    'seq_id': seq.seq_id,
                    'copy_number_in_sample': seq.copy_number
                })

    return sorted(
        sequences.values(),
        key=lambda v: v['copy_number_in_subject'],
        reverse=True
    )


def get_selection_pressure(session, clone_id):
    query = session.query(
        SelectionPressure
    ).filter(
        SelectionPressure.clone_id == clone_id,
    )

    pressure = {}
    for row in query:
        if row.sample_id is None:
            name = 'All'
        else:
            name = session.query(Sample.name).filter(
                Sample.id == row.sample_id).first().name
        if name not in pressure:
            pressure[name] = {
                'pressure': {},
                'sample': {
                    'id': row.sample_id,
                    'name': name
                }
            }
        pressure[name]['pressure'][row.threshold] = row.to_dict()

    return pressure.values()


def get_clone_tree(session, clone_id):
    tree = session.query(Clone.tree).filter(Clone.id == clone_id).first().tree
    return json.loads(tree) if tree is not None else None


def get_clone_overlap(session, sample_ids, filter_type, order_by='total_cnt',
                      paging=None):
    """Gets a list of clones and the samples in `samples` which they appear"""
    stats = session.query(
        CloneStats.clone_id,
        CloneStats.functional,
        CloneStats.unique_cnt,
        CloneStats.total_cnt
    ).filter(
        CloneStats.sample_id.in_(sample_ids),
    )
    if filter_type != 'clones_all':
        stats = stats.filter(
            CloneStats.functional == (filter_type == 'clones_functional')
        )

    aggregated_stats = {}
    for clone_id, functional, unique_cnt, total_cnt in stats:
        aggregated_stats.setdefault(clone_id, Counter()).update({
            'unique_cnt': unique_cnt,
            'total_cnt': total_cnt
        })
    aggregated_stats = sorted(
        aggregated_stats.items(),
        key=lambda kv: kv[1][order_by], reverse=True
    )

    if paging:
        start, per_page = paging
        aggregated_stats = aggregated_stats[
            (start - 1) * per_page:start * per_page]

    res = []
    for clone_id, counts in aggregated_stats:
        selected_samples = []
        other_samples = []
        clone = session.query(Clone).filter(Clone.id == clone_id).one()
        query = session.query(CloneStats).filter(
            CloneStats.clone_id == clone_id,
            ~CloneStats.sample_id.is_(None)
        ).order_by(
            desc(getattr(CloneStats, order_by))
        )
        for stat in query:
            data = {
                'id': stat.sample_id,
                'name': stat.sample.name,
                'unique_sequences': stat.unique_cnt,
                'total_sequences': stat.total_cnt
            }
            if stat.sample_id in sample_ids:
                selected_samples.append(data)
            else:
                other_samples.append(data)

        res.append({
            'unique_sequences': counts['unique_cnt'],
            'total_sequences': counts['total_cnt'],
            'clone': {
                'id': clone.id,
                'v_gene': clone.v_gene,
                'j_gene': clone.j_gene,
                'cdr3_num_nts': clone.cdr3_num_nts,
                'cdr3_aa': clone.cdr3_aa,
                'subject': _subject_to_dict(clone.subject),
            },
            'selected_samples': selected_samples,
            'other_samples': other_samples,
        })

    return res


def get_clones_in_samples(session, samples):
    return [e.id for e in session.query(
                   distinct(Sequence.clone_id).label('id')).filter(
                   Sequence.sample_id.in_(samples))]


def get_clones_in_subject(session, subject_id):
    return [e.id for e in session.query(Clone).filter(
        Clone.subject_id == subject_id)]


def get_grouping(sample, grouping):
    if grouping == 'sample':
        return sample.name
    elif grouping == 'subject':
        return sample.subject.identifier
    else:
        return sample.metadata_dict.get(grouping, None)


def get_v_usage(session, samples, filter_type, include_outliers,
                include_partials, grouping, by_family):
    """Gets the V-Gene usage percentages for samples"""
    if by_family:
        def name_func(s, prefix):
            return s.split('*')[0].split('-', 1)[0].replace(prefix, '')
    else:
        def name_func(s, prefix):
            return s.split('*')[0].replace(prefix, '')
    data = {}
    totals = {}
    prefix = ''
    for s in session.query(SampleStats)\
            .filter(SampleStats.filter_type == filter_type,
                    SampleStats.outliers == include_outliers,
                    SampleStats.full_reads != include_partials,
                    SampleStats.sample_id.in_(samples)):
        dist = json.loads(s.v_gene_dist)

        group_key = get_grouping(s.sample, grouping)

        if group_key not in data:
            data[group_key] = {}

        for v in dist:
            name, occ = v
            prefix = name[:4]
            ties = set([name_func(n, prefix) for n in name.split('|')])
            name = '|'.join(sorted(ties))

            if name not in data[group_key]:
                data[group_key][name] = 0
            data[group_key][name] += occ

    headers = []
    for group_key, names in data.items():
        totals[group_key] = float(sum(names.values()))
        for name, value in names.items():
            percent = round(100 * value / totals[group_key], 2)
            names[name] = percent
            if name not in headers and percent >= 1.0:
                headers.append(name)

    return data, sorted(headers), totals, prefix


def get_all_subjects(session, paging=None):
    subjects = []
    for subject in _page_query(
            session.query(Subject).order_by(Subject.identifier), paging):
        total_seqs = session.query(
            func.sum(SampleStats.sequence_cnt)
        ).filter(
            SampleStats.sample.has(subject=subject),
            SampleStats.filter_type == 'all',
            SampleStats.outliers == true(),
            SampleStats.full_reads == false()
        ).scalar()
        unique_seqs = session.query(
            func.sum(SampleStats.sequence_cnt)
        ).filter(
            SampleStats.sample.has(subject=subject),
            SampleStats.filter_type == 'unique',
            SampleStats.outliers == true(),
            SampleStats.full_reads == false()
        ).scalar()
        info = {
            'id': subject.id,
            'identifier': subject.identifier,
            'study': {
                'id': subject.study.id,
                'name': subject.study.name
            },
        }
        if total_seqs is not None:
            info['total_samples'] = session.query(
                func.count(Sample.id)
            ).filter(
                Sample.subject == subject
            ).scalar()
            info['total_seqs'] = int(total_seqs)
            info['unique_seqs'] = int(unique_seqs)
            info['total_clones'] = session.query(
                func.count(Clone.id)
            ).filter(
                Clone.subject_id == subject.id
            ).scalar()
        subjects.append(info)

    return subjects


def get_subject(session, sid):
    s = session.query(Subject).filter(Subject.id == sid).first()
    samples = get_samples(
        session,
        [e.id for e in session.query(Sample.id).filter(
            Sample.subject_id == sid)])

    subject = {
        'id': s.id,
        'identifier': s.identifier,
        'study': {
            'id': s.study.id,
            'name': s.study.name,
        },
        'samples': samples,
    }

    return subject


def analyze_samples(session, samples, filter_type, include_outliers,
                    include_partials, percentages, grouping):
    counts = {}

    stats = {}
    sample_info = {}
    dist_fields = [
        'v_match_dist', 'v_length_dist', 'v_identity_dist', 'j_match_dist',
        'j_length_dist', 'v_gene_dist', 'j_gene_dist', 'cdr3_length_dist',
        'copy_number_dist', 'quality_dist', 'sp_fwr_dist', 'sp_cdr_dist'
    ]
    cnt_fields = [
        'sequence_cnt', 'in_frame_cnt', 'stop_cnt', 'functional_cnt',
        'no_result_cnt'
    ]

    group_sizes = {}
    # Iterate over all filter types in the samples
    for stat in session.query(SampleStats).filter(
            SampleStats.sample_id.in_(samples),
            SampleStats.outliers == include_outliers,
            SampleStats.full_reads != include_partials):
        # Update the number of sequences in each filter
        if stat.filter_type not in counts:
            counts[stat.filter_type] = 0
        counts[stat.filter_type] += stat.sequence_cnt

        # If the sample is not already in the sample_info dictionary, add it
        if stat.sample.id not in sample_info:
            sample_info[stat.sample.id] = _sample_to_dict(stat.sample)

        if stat.filter_type == 'all':
            # If the filter is for all sequences, add the count fields to the
            # sample dictionary.
            for field in cnt_fields:
                sample_info[stat.sample.id][field] = getattr(stat, field)
            sample_info[stat.sample.id]['total_cnt'] = (
                sample_info[stat.sample.id]['sequence_cnt'] +
                sample_info[stat.sample.id]['no_result_cnt']
            )
        elif stat.filter_type == 'clones_all':
            # If it's the clones_all filter, update the total number of clones
            # in the samples
            sample_info[stat.sample_id].update({
                'clones_cnt': stat.sequence_cnt,
                'clones_functional_cnt': stat.functional_cnt
            })
        elif stat.filter_type == 'unique':
            # If it's the unique filter, update the total number of unique
            # sequences in the sample
            sample_info[stat.sample_id].update({
                'unique_cnt': stat.functional_cnt
            })
        elif stat.filter_type == 'unique_multiple':
            # If it's the unique_multiple filter, update the total number of
            # unique sequences that occur multiple times in the sample
            sample_info[stat.sample_id].update({
                'unique_multiple_cnt': stat.functional_cnt
            })

        # If this is the selected filter, group and tally the statistics
        if stat.filter_type == filter_type:
            group_key = get_grouping(stat.sample, grouping)

            if group_key not in stats:
                stats[group_key] = {}
                group_sizes[group_key] = 0
            group_sizes[group_key] += 1

            fields = _fields_to_dict(dist_fields, stat)

            for field, values in fields.items():
                if field not in stats[group_key]:
                    stats[group_key][field] = {}

                for (x, freq) in json.loads(values):
                    if x not in stats[group_key][field]:
                        stats[group_key][field][x] = 0
                    stats[group_key][field][x] += freq

    for group, key_dict in stats.items():
        for key, vals in key_dict.items():
            if key == 'quality_dist':
                vals = {k: v / group_sizes[group] for k, v in vals.items()}
            reduced = []
            for x in sorted(vals.keys()):
                val = vals[x]
                if (key != 'quality_dist' and
                        percentages and sum(vals.values()) > 0):
                    val = round(100 * vals[x] / sum(vals.values()), 2)
                reduced.append((x, val))
            key_dict[key] = reduced

    clone_cnts = {s.functional: s.clones for s in session.query(
        CloneStats.functional,
        func.count(distinct(CloneStats.clone_id)).label('clones')
    ).filter(
        CloneStats.sample_id.in_(samples)
    ).group_by(CloneStats.functional)}

    counts['clones_nonfunctional'] = clone_cnts.get(False, 0)
    counts['clones_functional'] = clone_cnts.get(True, 0)
    counts['clones_all'] = clone_cnts.get(False, 0) + clone_cnts.get(True, 0)

    return {'samples': sample_info, 'counts': counts, 'stats': stats}


def trace_seq_collapses(session, seq):
    collapse_info = session.query(
        SequenceCollapse
    ).filter(
        SequenceCollapse.sample_id == seq.sample_id,
        SequenceCollapse.seq_ai == seq.ai,
    ).first()

    if collapse_info is None:
        return None
    return {
        'sample_id': collapse_info.collapse_to_subject_sample_id,
        'sample_name': collapse_info.collapse_to_seq.sample.name,
        'ai': collapse_info.collapse_to_subject_seq_ai,
        'seq_id': collapse_info.collapse_to_subject_seq_id,
        'copy_number':
            collapse_info.collapse_to_seq.collapse.copy_number_in_subject,
        'instances':
            collapse_info.collapse_to_seq.collapse.instances_in_subject,
    }


def get_sequence(session, sample_id, seq_id):
    seq = session.query(Sequence)\
        .filter(Sequence.sample_id == sample_id,
                Sequence.seq_id == seq_id).one()
    ret = _fields_to_dict([
        'seq_id', 'partial', 'rev_comp', 'v_gene', 'j_gene', 'cdr3_nt',
        'cdr3_aa', 'germline', 'v_match', 'j_match', 'v_length',
        'j_length', 'in_frame', 'functional', 'stop', 'copy_number',
        'sequence', 'pre_cdr3_length', 'pre_cdr3_match', 'post_cdr3_length',
        'post_cdr3_match', 'seq_start', 'num_gaps',
        'probable_indel_or_misalign', 'insertions', 'deletions',
        'quality', 'regions'
    ], seq)
    ret['sample'] = _sample_to_dict(seq.sample)
    ret['read_start'] = re.compile(r'[N-]*').match(
        seq.sequence).span()[1] or 0

    ret['v_extent'] = seq.get_v_extent(in_clone=False)
    ret['mutations'] = {}
    if seq.mutations_from_clone:
        ret['mutations'] = json.loads(seq.mutations_from_clone)

    ret['clone'] = _clone_to_dict(seq.clone) if seq.clone is not None else None
    ret['collapse_info'] = trace_seq_collapses(session, seq)

    return ret


def get_sequences(session, filters, order_field, order_dir, subject_id=None,
                  paging=None):
    """Gets a list of all clones"""
    res = []
    query = session.query(Sequence).outerjoin(SequenceCollapse)

    copy_number_field = 'copy_number'
    if filters is not None:
        if filters.pop('copy_type', 'sample') == 'sample':
            copy_number_field = Sequence.copy_number
        else:
            copy_number_field = SequenceCollapse.copy_number_in_subject

        for key, value in filters.items():
            if value in [None, True, False]:
                continue
            value = str(value).strip()
            if len(value) > 0 and value is not None:
                if key == 'sample_id':
                    query = query.filter(Sequence.sample_id == int(value))
                elif key == 'in_frame':
                    query = query.filter(Sequence.in_frame == int(value))
                elif key == 'min_copy_number':
                    query = query.filter(copy_number_field >= int(value))
                elif key == 'max_copy_number':
                    query = query.filter(copy_number_field <= int(value))
                elif key == 'collapsed':
                    query = query.filter(copy_number_field > 0)
                else:
                    query = query.filter(getattr(Sequence, key).like(value))

    if subject_id is not None:
        query = query.filter(Sequence.subject_id == subject_id)

    if (filters is None or 'show_indel' not in filters or
            not filters['show_indel']):
        query = query.filter(Sequence.probable_indel_or_misalign == 0)

    if order_field is not None:
        if order_field == 'copy_number':
            order_field = copy_number_field
        else:
            order_field = getattr(Sequence, order_field)

        if order_dir == 'asc':
            query = query.order_by(order_field)
        else:
            query = query.order_by(desc(order_field))

    for row in _page_query(query, paging):
        fields = _fields_to_dict(
            ['seq_id', 'v_gene', 'j_gene', 'v_match', 'j_match',
             'v_length', 'j_length', 'cdr3_num_nts', 'cdr3_aa', 'in_frame',
             'functional', 'stop', 'partial', 'probable_indel_or_misalign',
             'copy_number'],
            row)
        if row.collapse is not None:
            fields.update({
                'copy_number_in_subject': row.collapse.copy_number_in_subject,
                'instances_in_subject': row.collapse.instances_in_subject
            })

        fields['sample'] = _sample_to_dict(row.sample)
        res.append(fields)

    return res



import json

import numpy as np

from sqlalchemy import and_, func

import immunedb.common.config as config
import immunedb.common.modification_log as mod_log
from immunedb.common.models import (Clone, NoResult, Sample, SampleStats,
                                    SelectionPressure, Sequence)
import immunedb.util.concurrent as concurrent
import immunedb.util.lookups as lookups
from immunedb.util.log import logger


_dist_fields = [
    'v_match',
    'j_match',
    'j_length',
    'v_gene',
    'j_gene',
    'copy_number',
    'v_length',
    'v_identity',
    'cdr3_length',
    'sp_fwr',
    'sp_cdr'
]

_seq_contexts = {
    'all': {
        'record_filter': lambda seq: True,
        'use_copy': True
    },
    'functional': {
        'record_filter': lambda seq: seq.functional,
        'use_copy': True
    },
    'nonfunctional': {
        'record_filter': lambda seq: not seq.functional,
        'use_copy': True
    },
    'unique': {
        'record_filter': lambda seq: seq.functional,
        'use_copy': False
    },
    'unique_multiple': {
        'record_filter': lambda seq: seq.functional and seq.copy_number > 1,
        'use_copy': False
    },
}


_clone_contexts = {
    'clones_all': {
        'record_filter': lambda functional: True,
    },
    'clones_functional': {
        'record_filter': lambda functional: functional
    },
    'clones_nonfunctional': {
        'record_filter': lambda functional: not functional
    },
}


class ContextStats(object):
    def __init__(self, record_filter):
        self._record_filter = record_filter
        self.distributions = {}

        for dist in _dist_fields:
            self.distributions[
                dist[0] if isinstance(dist, tuple) else dist] = {}

        self.sequence_cnt = 0
        self.in_frame_cnt = 0
        self.stop_cnt = 0
        self.functional_cnt = 0

    def _update(self, seq_record, in_frame, stop, functional, add):
        self.sequence_cnt += add
        if in_frame:
            self.in_frame_cnt += add
        if stop:
            self.stop_cnt += add
        if functional:
            self.functional_cnt += add

        for field in _dist_fields:
            try:
                value = getattr(seq_record, field)
                if value is None:
                    continue
            except AttributeError:
                continue
            if not isinstance(value, str):
                value = float(value)

            if field not in self.distributions:
                self.distributions[field] = {}
            if value not in self.distributions[field]:
                self.distributions[field][value] = 0

            self.distributions[field][value] += add


class SeqContextStats(ContextStats):
    def __init__(self, session, record_filter, use_copy):
        super(SeqContextStats, self).__init__(record_filter)
        self._session = session
        self._use_copy = use_copy
        self.quality = []

    def _update_quality(self, seq_record):
        if seq_record.quality is not None:
            diff = len(seq_record.quality) - len(self.quality)
            if diff > 0:
                self.quality.extend([[] for _ in range(0, diff)])

            for i, b in enumerate(seq_record.quality):
                if b != ' ':
                    self.quality[i].append(ord(b) - 33)

    def add_if_match(self, seq_record):
        if not self._record_filter(seq_record):
            return

        add = int(seq_record.copy_number) if self._use_copy else 1

        self._update(seq_record, seq_record.in_frame, seq_record.stop,
                     seq_record.functional, add)
        self._update_quality(seq_record)


class CloneContextStats(ContextStats):
    def __init__(self, record_filter, seqs):
        super(CloneContextStats, self).__init__(record_filter)
        self._seqs = seqs

    def add_if_match(self, clone, in_frame, stop, functional):
        if not self._record_filter(functional):
            return

        self._update(clone, in_frame, stop, functional, 1)


class SampleStatsWorker(concurrent.Worker):
    def __init__(self, session):
        self._session = session

    def do_task(self, args):
        if args['func'] == 'seq':
            func = self._calculate_seq_stats
        elif args['func'] == 'clone':
            func = self._calculate_clone_stats
        self.info(('Processing {} for sample {}, include_outliers {}, '
                   'only_full_reads {}').format(
                   'sequences' if args['func'] == 'seq' else 'clones',
                   args['sample_id'], args['include_outliers'],
                   args['only_full_reads']))
        func(args['sample_id'], args['min_cdr3'], args['max_cdr3'],
             args['include_outliers'], args['only_full_reads'])
        self._session.commit()

    def _add_stat(self, stats, sample_id, include_outliers,
                  only_full_reads):
        for name, stat in stats.items():
            ss = SampleStats(
                sample_id=sample_id,
                filter_type=name,
                outliers=include_outliers,
                full_reads=only_full_reads,
                sequence_cnt=stat.sequence_cnt,
                in_frame_cnt=stat.in_frame_cnt,
                stop_cnt=stat.stop_cnt,
                functional_cnt=stat.functional_cnt,
                no_result_cnt=self._session.query(NoResult).filter(
                    NoResult.sample_id == sample_id).count(),
            )

            if hasattr(stat, 'quality'):
                quality_dist = []
                for pos, quals in enumerate(stat.quality):
                    if len(quals) > 0:
                        quality_dist.append((pos, round(np.mean(quals), 2)))
                ss.quality_dist = json.dumps(quality_dist)
            else:
                ss.quality_dist = json.dumps([])

            for dname, dist in stat.distributions.items():
                setattr(ss, '{}_dist'.format(dname),
                        json.dumps([(k, v) for k, v in dist.items()]))
            self._session.add(ss)

    def _calculate_seq_stats(self, sample_id, min_cdr3, max_cdr3,
                             include_outliers, only_full_reads):
        seq_statistics = {}
        for name, stat in _seq_contexts.items():
            seq_statistics[name] = SeqContextStats(self._session, **stat)

        # TODO: This should be automatically generated from _dist_fields
        query = self._session.query(
            Sequence.quality,
            Sequence.sample_id,
            Sequence.v_match,
            Sequence.j_match,
            Sequence.j_length,
            Sequence.v_gene,
            Sequence.j_gene,
            Sequence.in_frame,
            Sequence.stop,
            Sequence.functional,
            Sequence.copy_number,
            (Sequence.v_length + Sequence.num_gaps).label('v_length'),
            (
                func.ceil(100 * Sequence.v_match / Sequence.v_length)
            ).label('v_identity'),
            Sequence.cdr3_num_nts.label('cdr3_length'),
        ).filter(
            Sequence.sample_id == sample_id
        )

        if not include_outliers and min_cdr3 is not None:
            query = query.filter(Sequence.cdr3_num_nts >= min_cdr3,
                                 Sequence.cdr3_num_nts <= max_cdr3)
        if only_full_reads:
            query = query.filter(Sequence.partial == 0)

        for seq in query:
            for name, stat in seq_statistics.items():
                stat.add_if_match(seq)

        self._add_stat(seq_statistics, sample_id, include_outliers,
                       only_full_reads)

    def _calculate_clone_stats(self, sample_id, min_cdr3, max_cdr3,
                               include_outliers, only_full_reads):
        clone_statistics = {}
        for name, stat in _clone_contexts.items():
            clone_statistics[name] = CloneContextStats(seqs=None, **stat)

        # TODO: This should be automatically generated from _dist_fields
        query = self._session.query(
            Sequence.clone_id,
            func.round(func.avg(Sequence.v_match)).label('v_match'),
            func.round(func.avg(Sequence.j_match)).label('j_match'),
            func.round(func.avg(Sequence.j_length)).label('j_length'),
            Sequence.v_gene,
            Sequence.j_gene,
            func.count(Sequence.seq_id).label('copy_number'),
            func.round(
                func.avg(Sequence.v_length + Sequence.num_gaps)
            ).label('v_length'),
            func.round(
                func.avg(100 * Sequence.v_match / Sequence.v_length)
            ).label('v_identity'),
            Sequence.cdr3_num_nts.label('cdr3_length'),
            SelectionPressure.sigma_fwr.label('sp_fwr'),
            SelectionPressure.sigma_cdr.label('sp_cdr'),
        ).join(
            SelectionPressure,
            and_(
                SelectionPressure.clone_id == Sequence.clone_id,
                SelectionPressure.sample_id == Sequence.sample_id
            ),
            isouter=True
        ).filter(
            Sequence.sample_id == sample_id,
            ~Sequence.clone_id.is_(None)
        )

        if only_full_reads:
            query = query.filter(Sequence.partial == 0)
        query = query.group_by(Sequence.clone_id)

        for clone in query:
            clone_info = self._session.query(Clone.cdr3_nt).filter(
                Clone.id == clone.clone_id).first()
            in_frame = len(clone_info.cdr3_nt) % 3 == 0
            stop = '*' in lookups.aas_from_nts(clone_info.cdr3_nt)
            functional = in_frame and not stop
            for name, stat in clone_statistics.items():
                stat.add_if_match(clone, in_frame, stop, functional)

        self._add_stat(clone_statistics, sample_id, include_outliers,
                       only_full_reads)


def _get_cdr3_bounds(session, sample_id):
    cdr3_fld = Sequence.cdr3_num_nts
    cdr3s = []

    query = session.query(
        func.sum(Sequence.copy_number).label('copy_number'),
        cdr3_fld.label('cdr3_len')
    ).filter(
        Sequence.sample_id == sample_id,
        Sequence.copy_number > 1,
        Sequence.functional == 1
    ).group_by(cdr3_fld)

    for seq in query:
        cdr3s += [seq.cdr3_len] * int(seq.copy_number)
    if len(cdr3s) == 0:
        return None, None
    q25, q75 = np.percentile(cdr3s, [25, 75])
    iqr = q75 - q25
    return float(q25 - 1.5 * iqr), float(q75 + 1.5 * iqr)


def _queue_tasks(session, sample_id, force, tasks):
    logger.info('Creating task queue to generate stats for sample {}.'.format(
        sample_id))
    existing_seq = session.query(Sequence).filter(
        Sequence.sample_id == sample_id)
    existing_nores = session.query(NoResult).filter(
        NoResult.sample_id == sample_id)
    if existing_seq.first() is None and existing_nores.first() is None:
        logger.warning('\tSKIPPING since there are no sequences in the '
                       'sample')
        return

    existing = session.query(SampleStats.sample_id).filter(
        SampleStats.sample_id == sample_id).first() is not None
    if force and existing:
        logger.warning('\tFORCING regeneration of stats')
    elif not force and existing:
        logger.warning('\tSKIPPING stats since they already exists and the '
                       '--force flag was not specified.')
        return

    min_cdr3, max_cdr3 = _get_cdr3_bounds(session, sample_id)
    for include_outliers in [True, False]:
        for only_full_reads in [True, False]:
            tasks.add_task({
                'func': 'seq',
                'sample_id': sample_id,
                'min_cdr3': min_cdr3,
                'max_cdr3': max_cdr3,
                'include_outliers': include_outliers,
                'only_full_reads': only_full_reads
            })
            tasks.add_task({
                'func': 'clone',
                'sample_id': sample_id,
                'min_cdr3': min_cdr3,
                'max_cdr3': max_cdr3,
                'include_outliers': include_outliers,
                'only_full_reads': only_full_reads
            })


def run_sample_stats(session, args):
    np.seterr(all='raise')
    mod_log.make_mod('sample_stats', session=session, commit=True,
                     info=vars(args))

    if args.sample_ids is None:
        samples = [s.id for s in session.query(Sample.id)]
    else:
        samples = args.sample_ids

    if args.force:
        q = session.query(SampleStats).filter(
            SampleStats.sample_id.in_(samples))
        q.delete(synchronize_session=False)
        session.commit()

    tasks = concurrent.TaskQueue()
    for sample_id in samples:
        _queue_tasks(session, sample_id, args.force, tasks)

    for i in range(0, args.nproc):
        session = config.init_db(args.db_config)
        tasks.add_worker(SampleStatsWorker(session))

    tasks.start()
    session.commit()
    session.close()



from collections import OrderedDict

from sqlalchemy import desc
from sqlalchemy.sql import text

import numpy as np
import pandas as pd
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import fcluster, linkage

import dnautils
from immunedb.common.models import (CDR3_OFFSET, Clone, Sequence,
                                    SequenceCollapse, Subject)
import immunedb.common.modification_log as mod_log
import immunedb.common.config as config
import immunedb.util.concurrent as concurrent
import immunedb.util.funcs as funcs
import immunedb.util.lookups as lookups
from immunedb.util.log import logger


def generate_consensus(session, clone_ids):
    """Generates consensus CDR3s for clones.

    :param Session session: The database session
    :param list clone_ids: The list of clone IDs to assign to groups

    """

    if not clone_ids:
        return
    for clone in funcs.periodic_commit(
            session,
            session.query(Clone).filter(Clone.id.in_(clone_ids)),
            interval=1000):
        seqs = session.query(
            Sequence
        ).join(SequenceCollapse).filter(
            Sequence.clone_id == clone.id,
            SequenceCollapse.copy_number_in_subject > 0
        ).all()
        clone.cdr3_nt = funcs.consensus([s.cdr3_nt for s in seqs])
        clone.cdr3_aa = lookups.aas_from_nts(clone.cdr3_nt)

        clone.germline = generate_germline(session, seqs, clone)

        clone.overall_total_cnt = sum([s.copy_number for s in seqs])

        clone.functional = (
            clone.cdr3_num_nts % 3 == 0 and
            '*' not in clone.cdr3_aa and
            not lookups.has_stop(clone.germline)
        )

    session.commit()


def generate_germline(session, seqs, clone):
    rep_seq = seqs[0]
    germ = rep_seq.alignment_without_insertions[0]
    return ''.join((
        germ[:CDR3_OFFSET],
        '-' * clone.cdr3_num_nts,
        germ[CDR3_OFFSET + clone.cdr3_num_nts:]
    ))


def push_clone_ids(session):
    session.connection(mapper=Sequence).execute(text('''
        UPDATE
            sequences AS s
        JOIN sequence_collapse AS c
            ON s.sample_id=c.sample_id AND s.ai=c.seq_ai
        JOIN sequences as s2
            ON c.collapse_to_subject_seq_ai=s2.ai
        SET s.clone_id=s2.clone_id
        WHERE s.seq_id!=s2.seq_id
    '''))


def collapse_similar_cdr3s(session, buckets, difference_allowed):
    logger.info('Collapsing similar clones in {} buckets'.format(
        buckets.count()
    ))
    for i, bucket in enumerate(buckets):
        clones = session.query(
            Clone.id, Clone.cdr3_aa, Clone.cdr3_nt
        ).filter(
            Clone.subject_id == bucket.subject_id,
            Clone.cdr3_num_nts == bucket.cdr3_num_nts,
        ).order_by(
            Clone.overall_total_cnt.desc()
        )
        if clones.count() < 2:
            continue
        logger.info('Reducing bucket {} / {} ({} clones)'.format(
            i, buckets.count(), clones.count()))
        reduced = {}
        for c in clones:
            for larger_cdr3_nt, others in reduced.items():
                if (dnautils.hamming(larger_cdr3_nt, c.cdr3_nt) <=
                        difference_allowed):
                    others.append(c.id)
                    break
            else:
                reduced[c.cdr3_nt] = [c.id]

        for collapse in reduced.values():
            rep_id, others = collapse[0], collapse[1:]
            session.query(Sequence).filter(
                Sequence.clone_id.in_(others)
            ).update(
                {'clone_id': rep_id},
                synchronize_session=False
            )
            session.query(Clone).filter(
                Clone.id.in_(others)
            ).delete(synchronize_session=False)
    session.commit()


def similar_to_all(seq, rest, field, min_similarity):
    """Determines if the string ``seq`` is at least ``min_similarity``
    similar to the list of strings ``rest``.

    :param str seq: The string to compare
    :param list rest: The list of strings to compare to
    :param int min_similarity: Minimum fraction to be considered similar

    :returns: If ``seq`` is similar to every sequence in ``rest``
    :rtype: bool

    """
    for comp_seq in rest:
        dist = dnautils.hamming(
            getattr(comp_seq, 'cdr3_' + field).replace('X', '-'),
            getattr(seq, 'cdr3_' + field).replace('X', '-')
        )
        sim_frac = 1 - dist / len(comp_seq.cdr3_aa)
        if sim_frac < min_similarity:
            return False
    return True


def can_subclone(sub_clone, parent_clone):
    return dnautils.equal(
        parent_clone.cdr3_aa.replace('X', '-'),
        sub_clone.cdr3_aa.replace('X', '-')
    )


class ClonalWorker(concurrent.Worker):
    defaults = {
        # common
        'min_copy': 2,
        'max_padding': None,
        'min_similarity': 0.85,
    }

    def __init__(self, session, **kwargs):
        self.session = session
        for prop, default in self.defaults.items():
            setattr(self, prop, kwargs.get(prop, default))
        for prop, value in kwargs.items():
            if prop not in self.defaults:
                setattr(self, prop, value)
        self._tasks = 0

    def get_bucket_seqs(self, bucket, sort):
        query = self.session.query(
            Sequence
        ).join(SequenceCollapse).filter(
            Sequence.subject_id == bucket.subject_id,
            Sequence.v_gene == bucket.v_gene,
            Sequence.j_gene == bucket.j_gene,
            Sequence.cdr3_num_nts == bucket.cdr3_num_nts,

            SequenceCollapse.copy_number_in_subject >= self.min_copy,
        )
        if sort:
            query = query.order_by(
                desc(SequenceCollapse.copy_number_in_subject),
                Sequence.ai
            )
        if self.max_padding is not None:
            query = query.filter(Sequence.seq_start <= self.max_padding)
        return query

    def do_task(self, bucket):
        self.run_bucket(bucket)
        self._tasks += 1
        if self._tasks % 100 == 0:
            self.session.commit()
            self.info('Collapsed {} buckets'.format(self._tasks))

    def cleanup(self):
        self.session.commit()
        self.session.close()


class SimilarityClonalWorker(ClonalWorker):
    def run_bucket(self, bucket):
        clones = OrderedDict()
        consensus_needed = set([])
        query = self.get_bucket_seqs(bucket, sort=True)

        if query.count() > 0:
            for seq in query:
                if seq.clone_id not in clones:
                    clones[seq.clone_id] = []
                clones[seq.clone_id].append(seq)
            if None in clones:
                for seq_to_add in clones[None]:
                    for clone_id, existing_seqs in clones.items():
                        if clone_id is None:
                            continue
                        if similar_to_all(seq_to_add, existing_seqs,
                                          self.level, self.min_similarity):
                            existing_seqs.append(seq_to_add)
                            break
                    else:
                        new_clone = Clone(subject_id=seq.subject_id,
                                          v_gene=seq.v_gene,
                                          j_gene=seq.j_gene,
                                          cdr3_num_nts=seq.cdr3_num_nts)
                        self.session.add(new_clone)
                        self.session.flush()
                        clones[new_clone.id] = [seq_to_add]
                del clones[None]

            for clone_id, seqs in clones.items():
                to_update = [
                    {
                        'sample_id': s.sample_id,
                        'ai': s.ai,
                        'clone_id': clone_id
                    } for s in seqs if s.clone_id is None
                ]
                if len(to_update) > 0:
                    self.session.bulk_update_mappings(Sequence, to_update)
                    consensus_needed.add(clone_id)
        generate_consensus(self.session, consensus_needed)


class ClusteringClonalWorker(ClonalWorker):
    def get_distances(self, seqs):
        dists = np.zeros((len(seqs), len(seqs)))
        for i, s1 in enumerate(seqs):
            for j, s2 in enumerate(seqs):
                dists[i, j] = dists[j, i] = dnautils.hamming(s1, s2) / len(s1)
        return dists

    def assign_clones(self, df):
        seq = df.iloc[0]
        new_clone = Clone(subject_id=int(seq.subject_id),
                          v_gene=seq.v_gene,
                          j_gene=seq.j_gene,
                          cdr3_num_nts=int(seq.cdr3_num_nts))
        self.session.add(new_clone)
        self.session.flush()
        to_update = [
            {
                'sample_id': s.sample_id,
                'ai': s.ai,
                'clone_id': new_clone.id
            } for _, s in df.iterrows()
        ]
        self.session.bulk_update_mappings(Sequence, to_update)
        return int(new_clone.id)

    def run_bucket(self, bucket):
        seqs = self.get_bucket_seqs(bucket, sort=True)
        if seqs.count():
            df = pd.DataFrame({
                'sample_id': s.sample_id,
                'ai': s.ai,

                'subject_id': s.subject_id,
                'v_gene': s.v_gene,
                'j_gene': s.j_gene,
                'cdr3_num_nts': s.cdr3_num_nts,

                'cdr3': getattr(s, 'cdr3_' + self.level).replace('X', '-'),
                'current_clone_id': s.clone_id
            } for s in seqs)

            if len(df) == 1:
                df['clone'] = 1
            else:
                dists = squareform(self.get_distances(df.cdr3))
                clusters = fcluster(
                    linkage(dists, 'complete'),
                    1 - self.min_similarity,
                    criterion='distance'
                )
                df['clone'] = clusters
            # NOTE: .groupby() doesn't work here since self.assign_clones has
            # side effects
            consensus_needed = []
            for clone in sorted(df.clone.unique()):
                new_id = self.assign_clones(df[df.clone == clone])
                consensus_needed.append(new_id)
            generate_consensus(self.session, consensus_needed)


class SubcloneWorker(concurrent.Worker):
    def __init__(self, session):
        self.session = session

    def do_task(self, bucket):
        clones = self.session.query(Clone).filter(
            Clone.subject_id == bucket.subject_id,
            Clone.v_gene == bucket.v_gene,
            Clone.j_gene == bucket.j_gene,
            Clone.cdr3_num_nts == bucket.cdr3_num_nts,
        ).all()

        if not clones:
            return
        # The clones with indels are the only ones which can be subclones
        parent_clones = set([c for c in clones if len(c.insertions) == 0 and
                             len(c.deletions) == 0])
        potential_subclones = [c for c in clones if c not in parent_clones and
                               c.parent_id is None]
        self.info('Bucket {} has {} clones; parents={}, subs={}'.format(
            bucket, len(clones), len(parent_clones), len(potential_subclones)))

        for subclone in potential_subclones:
            for parent in parent_clones:
                if can_subclone(subclone, parent):
                    subclone.parent = parent
                    break
        self.session.commit()

    def cleanup(self):
        self.session.commit()
        self.session.close()


def run_subclones(session, subject_ids, args):
    tasks = concurrent.TaskQueue()
    for subject_id in subject_ids:
        logger.info('Generating subclone task queue for subject {}'.format(
            subject_id))
        buckets = session.query(
            Clone.subject_id, Clone.v_gene, Clone.j_gene, Clone.cdr3_num_nts
        ).filter(
            Clone.subject_id == subject_id
        ).group_by(
            Clone.subject_id, Clone.v_gene, Clone.j_gene, Clone.cdr3_num_nts
        )
        for bucket in buckets:
            tasks.add_task(bucket)

    logger.info('Generated {} total subclone tasks'.format(tasks.num_tasks()))
    for i in range(0, min(tasks.num_tasks(), args.nproc)):
        tasks.add_worker(SubcloneWorker(config.init_db(args.db_config)))
    tasks.start()


def run_clones(session, args):
    """Runs the clone-assignment pipeline stage.

    :param Session session: The database session
    :param Namespace args: The arguments passed to the command

    """
    if args.subject_ids is None:
        subject_ids = [s.id for s in session.query(Subject.id)]
    else:
        subject_ids = args.subject_ids
    mod_log.make_mod('clones', session=session, commit=True, info=vars(args))

    if not args.skip_regen:
        logger.info('Deleting existing clones')
        q = session.query(Clone).filter(
            Clone.subject_id.in_(subject_ids)
        )
        if args.gene:
            q = q.filter(Clone.v_gene.like(args.gene + '%'))
        q.delete(synchronize_session=False)
        session.commit()

    tasks = concurrent.TaskQueue()
    all_buckets = []
    for subject_id in subject_ids:
        logger.info('Generating task queue for subject {}'.format(
            subject_id))
        buckets = session.query(
            Sequence.subject_id, Sequence.v_gene, Sequence.j_gene,
            Sequence.cdr3_num_nts
        ).filter(
            Sequence.subject_id == subject_id,
            Sequence.clone_id.is_(None)
        ).group_by(
            Sequence.subject_id, Sequence.v_gene, Sequence.j_gene,
            Sequence.cdr3_num_nts
        )
        for bucket in buckets:
            if not args.gene or bucket.v_gene.startswith(args.gene):
                tasks.add_task(bucket)
        all_buckets.extend(buckets)

    logger.info('Generated {} total tasks'.format(tasks.num_tasks()))

    methods = {
        'similarity': SimilarityClonalWorker,
        'cluster': ClusteringClonalWorker
    }
    for i in range(0, min(tasks.num_tasks(), args.nproc)):
        worker = methods[args.method](
            config.init_db(args.db_config), **args.__dict__
        )
        tasks.add_worker(worker)
    tasks.start()

    session.commit()
    if args.reduce_difference:
        buckets = session.query(
            Clone.subject_id,
            Clone.cdr3_num_nts
        ).filter(
            Clone.subject_id.in_(subject_ids)
        ).group_by(
            Clone.subject_id,
            Clone.cdr3_num_nts
        )
        collapse_similar_cdr3s(session, buckets, args.reduce_difference)
    else:
        logger.info('Skipping reduce since --reduce-differece set to 0')

    push_clone_ids(session)
    session.commit()

    if not args.skip_subclones:
        run_subclones(session, subject_ids, args)
    else:
        logger.info('Skipping subclones')



from collections import Counter

import numpy as np
from sqlalchemy.orm import joinedload

from immunedb.common.models import CloneStats, Sample
from immunedb.exporting.tsv_writer import StreamingTSV
from immunedb.exporting.writer import ExportWriter
from immunedb.util.funcs import chunks
from immunedb.util.lookups import aas_from_nts
from immunedb.util.log import logger

DEFAULT_CLONE_FIELDS = [
    'clone_id', 'subject', 'v_gene', 'j_gene', 'functional', 'insertions',
    'deletions', 'cdr3_nt', 'cdr3_num_nts', 'cdr3_aa',
    'uniques', 'instances', 'copies', 'germline', 'parent_id',
    'avg_v_identity', 'top_copy_seq'
]


def get_clone_row(clone):
    row = {}
    for field in DEFAULT_CLONE_FIELDS:
        try:
            row[field] = getattr(clone, field)
        except AttributeError:
            pass
    row.update({
        'clone_id': clone.id,
        'subject': clone.subject.identifier,
        'functional': 'T' if clone.functional else 'F',
        'insertions': clone._insertions,
        'deletions': clone._deletions,
        'uniques': clone.overall_unique_cnt,
        'instances': clone.overall_instance_cnt,
        'copies': clone.overall_total_cnt,
    })
    return row


def get_immunedb_output(session, clones):
    writer = StreamingTSV(DEFAULT_CLONE_FIELDS)
    yield writer.writeheader()

    for clone, agg in clones.items():
        counts = agg['counts']
        row = get_clone_row(clone)
        row['copies'] = counts['copies']
        row['instances'] = counts['instances']
        row['top_copy_seq'] = agg['top_seq']
        row['avg_v_identity'] = round(agg['avg_v_identity'], 4)
        yield writer.writerow(row)


def get_vdjtools_output(session, clones):
    writer = StreamingTSV(['count', 'freq', 'cdr3nt', 'cdr3aa', 'v', 'd', 'j'])
    counts = Counter()
    total_copies = 0
    for clone, agg in clones.items():
        key = (clone.v_gene, clone.j_gene, clone.cdr3_nt)
        counts[key] += agg['counts']['copies']
        total_copies += counts[key]

    yield writer.writeheader()
    for key in sorted(counts, key=counts.get, reverse=True):
        count = counts[key]
        v, j, cdr3_nt = key
        yield writer.writerow({
            'count': count,
            'freq': count / total_copies,
            'cdr3nt': cdr3_nt,
            'cdr3aa': aas_from_nts(cdr3_nt),
            'v': v,
            'd': '.',
            'j': j,
        })


def _get_feature(stat, feature):
    if feature == 'subject':
        return stat.clone.subject.identifier
    elif feature == 'sample':
        return stat.sample.name
    else:
        return stat.sample.metadata_dict.get(feature, 'NA')


def get_pooled_samples(session, sample_ids, features):
    stats = session.query(
        CloneStats
    ).options(
        joinedload(CloneStats.sample),
        joinedload(CloneStats.clone).defer('tree'),
    )
    aggregated = {}
    for chunk_sample_ids in chunks(sample_ids, 20):
        for stat in stats.filter(CloneStats.sample_id.in_(chunk_sample_ids)):
            sample_feature = tuple(
                _get_feature(stat, f) for f in sorted(features)
            )
            key = (stat.sample.subject.identifier, sample_feature)
            agg = aggregated.setdefault(key, {}).setdefault(
                stat.clone,
                {'top_seq': None, 'top_copies': 0, 'counts': Counter(),
                    'avg_v_identity': []}
            )
            if stat.top_copy_seq_copies > agg['top_copies']:
                agg['top_copies'] = stat.top_copy_seq_copies
                agg['top_seq'] = stat.top_copy_seq_sequence
            agg['counts']['copies'] += stat.total_cnt
            agg['counts']['instances'] += stat.unique_cnt
            agg['avg_v_identity'].append(stat.avg_v_identity * stat.total_cnt)

    for clones in aggregated.values():
        for agg in clones.values():
            agg['avg_v_identity'] = np.sum(
                    agg['avg_v_identity']) / agg['counts']['copies']

    return aggregated


def get_filename(subject, feature_keys, feature_values):
    if feature_keys == ('subject',):
        return '{}.pooled.tsv'.format(subject)
    feature_value = '_AND_'.join(feature_values)
    return '{}.{}.pooled.tsv'.format(subject, feature_value)


def write_pooled_clones(session, out_format, sample_ids=None,
                        pool_on=('sample',), zipped=False, **kwargs):
    # Samples and subjects can't be combined with other features
    exclusives = set(pool_on).intersection(set(('sample', 'subject')))
    if len(pool_on) > 1 and exclusives:
        pool_on = (list(exclusives)[0],)
        logger.warning('You specified pooling on {feat} which '
                       'cannot be combined with other features.'
                       '  Using only {feat}.'.format(feat=pool_on[0]))

    logger.info('Writing clones pooled by {} in {} format'.format(
        ','.join(pool_on), out_format))

    sample_ids = sample_ids or [s.id for s in session.query(Sample)]
    aggregated = get_pooled_samples(session, sample_ids, pool_on)

    output_func = {
        'immunedb': get_immunedb_output,
        'vdjtools': get_vdjtools_output
    }[out_format]
    with ExportWriter(zipped=zipped) as fh:
        for (subject, feature_value), clones in aggregated.items():
            logger.info('Pooling subject {} for feature(s) {}'.format(
                subject,
                ','.join(feature_value)))
            fh.set_filename(get_filename(subject, pool_on, feature_value))
            fh.write(output_func(session, clones))
        return fh.get_zip_value()



from collections import OrderedDict
import csv
import re

from immunedb.common.models import Sample, Sequence
from immunedb.util.log import logger


def _get_entries(seq, inference, gene_db):
    entry = OrderedDict()
    trim_start = re.search('[N-]*', seq.sequence).end()
    trim_end = re.search('[N-]*', seq.sequence[::-1]).end()
    trimmed_seq = seq.sequence[trim_start:len(seq.sequence) - trim_end]
    # Offsets of regions/segments
    v_gaps = trimmed_seq.count('-')
    trimmed_seq = trimmed_seq.replace('-', '')
    v_region = (1, len(trimmed_seq), 'V_region')
    v_segment = (1, 309 - v_gaps - trim_start, 'V_segment')
    cds = (
        v_segment[1] + 1,
        v_segment[1] + seq.cdr3_num_nts,
        'CDS'
    )
    j_segment = (cds[1] + 1, v_region[1], 'J_segment')

    # Key
    entry[('>Feature', seq.seq_id.split(' ')[0])] = []
    # V region
    entry[v_region] = []
    # V segment
    first_v = seq.v_gene.split('|')[0]
    entry[v_segment] = (
        ('gene', first_v),
        ('db_xref', '{}:{}'.format(gene_db, first_v)),
        ('inference', inference),
    )

    # J segment
    first_j = seq.j_gene.split('|')[0]
    entry[j_segment] = (
        ('gene', first_j),
        ('db_xref', '{}:{}'.format(gene_db, first_j)),
        ('inference', inference),
    )

    # CDS
    cds = ('<{}'.format(cds[0]), '>{}'.format(cds[1]), 'CDS')
    entry[cds] = (
        ('function', 'junction'),
        ('codon_start', 1),
        ('product', 'immunoglobulin heavy chain variable domain'),
        ('inference', inference)
    )

    return entry, trimmed_seq


def _write_sample(session, sample_id, gene_db, inference, header):
    for sample in session.query(Sample):
        seqs = session.query(
            Sequence.seq_id,
            Sequence.v_gene,
            Sequence.j_gene,
            Sequence.cdr3_num_nts,
            Sequence.sequence,
            Sequence.copy_number
        ).filter(
            Sequence.sample_id == sample_id
        ).filter(Sequence.stop == 0)
        with open('{}.tbl'.format(sample.name), 'w+') as gb_fh:
            writer = csv.writer(gb_fh, delimiter='\t')
            with open('{}.fsa'.format(sample.name), 'w+') as fasta_fh:
                for seq in seqs:
                    gb_entry, fasta_seq = _get_entries(
                        seq, inference, gene_db)
                    for entry, indented in gb_entry.items():
                        writer.writerow(entry)
                        if indented:
                            for indent in indented:
                                writer.writerow(('', '', '') + indent)
                    seq_header = header + ' [note=AIRR_READ_COUNT:{}]'.format(
                        seq.copy_number)
                    fasta_fh.write('>{}\n{}\n'.format(
                        seq.seq_id.split(' ')[0] + ' {}'.format(seq_header),
                        fasta_seq))


def write_genbank(session, args):
    args.inference = 'alignment:' + args.inference

    header = (
        '[organism={}] '
        '[moltype={}] '
        '[keywords=AIRR]').format(
        args.species, args.mol_type)

    for sample in session.query(Sample):
        logger.info('Exporting sample {}'.format(sample.name))
        _write_sample(session, sample.id, args.gene_db, args.inference,
                      header)



import io
import zipfile


class ExportWriter:
    def __init__(self, zipped=False):
        self.zipped = zipped
        self.filename = None

    def get_handle(self, fn, mode='w+'):
        class WriterFH:
            def __init__(self, writer):
                self.writer = writer

            def __enter__(self):
                self.buffer = io.BytesIO() if 'b' in mode else io.StringIO()
                return self.buffer

            def __exit__(self, exc_type, exc_value, traceback):
                self.writer.set_filename(fn, mode)
                self.writer.write(self.buffer.getvalue())

        return WriterFH(self)

    def set_filename(self, fn, mode='w+'):
        if self.zipped:
            self.filename = fn
        else:
            self.fh = open(fn, mode)

    def write(self, data):
        if self.zipped:
            self.zip.writestr(
                self.filename,
                data if type(data) == bytes else ''.join(data)
            )
        else:
            if type(data) == bytes:
                self.fh.write(data)
            else:
                for chunk in data:
                    self.fh.write(chunk)

    def get_zip_value(self, write_fn=None):
        if self.zipped:
            self.zip.close()
            val = self.out.getvalue()
            if write_fn:
                with open(write_fn, 'wb+') as fh:
                    fh.write(val)
            else:
                return val

    def __enter__(self):
        if self.zipped:
            self.out = io.BytesIO()
            self.zip = zipfile.ZipFile(self.out, 'w')
        else:
            self.filename = None
            self.fh = None
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.zipped:
            self.zip.close()
        else:
            if self.fh:
                self.fh.close()



import os

from setuptools import setup, Extension

if os.environ.get('READTHEDOCS'):
    install_requires = ['sphinxcontrib-websupport']
    ext_modules = []
else:
    with open('requirements.txt') as req:
        install_requires = [l.strip() for l in req]
    ext_modules = [
        Extension('dnautils', sources=['lib/dnautils.c'],
                  extra_compile_args=['-std=c99'])
    ]

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='ImmuneDB',
    version='0.29.4',
    author='Aaron M. Rosenfeld',
    author_email='aaron.rosenfeld@pennmedicine.upenn.edu',
    url='https://github.com/arosenfeld/immunedb',
    packages=[
        'immunedb',
        'immunedb.aggregation',
        'immunedb.api',
        'immunedb.common',
        'immunedb.exporting',
        'immunedb.exporting.clones',
        'immunedb.identification',
        'immunedb.importing',
        'immunedb.trees',
        'immunedb.util',
    ],
    scripts=[
        'bin/immunedb_admin',
        'bin/immunedb_clones',
        'bin/immunedb_clone_import',
        'bin/immunedb_clone_stats',
        'bin/immunedb_clone_pressure',
        'bin/immunedb_clone_trees',
        'bin/immunedb_collapse',
        'bin/immunedb_export',
        'bin/immunedb_local_align',
        'bin/immunedb_identify',
        'bin/immunedb_import',
        'bin/immunedb_metadata',
        'bin/immunedb_modify',
        'bin/immunedb_rest',
        'bin/immunedb_sample_stats',
        'bin/immunedb_sql',
        'bin/immunedb_genotype',
        'bin/run_tigger'
    ],
    install_requires=install_requires,
    ext_modules=ext_modules,
    license='LICENSE.txt',
    description='''A system for the analysis and exploration of high-throughput
        adaptive immune receptor sequencing data''',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ]
)
